from __future__ import annotations

import geopandas as gpd
import numpy as np
import xarray as xr


# TODO: better to split to separate functions, each with unique return type
def get_elev_diff(
    source: xr.Dataset | gpd.GeoDataFrame,
    reference: xr.DataArray,
    source_col: str | None = None,
    diff_col_name: str | None = "elev_diff",
) -> xr.Dataset | gpd.GeoDataFrame:
    """
    Calculate elevation differences between source and reference datasets.
    If the source is an xr.Dataset, this assumes the source dataset will
    have elevation values stored in variable name 'elevation'.

    Parameters
    ----------
    source: xr.Dataset | gpd.GeoDataFrame
        Source elevation data, either as raster or points
    reference: xr.DataArray
        Reference elevation data as raster
    source_col: str | None
        Column name containing elevation values to difference if source is GeoDataFrame (e.g. h_li for ICESat-2 ATL06)
    diff_col_name: str
        Resulting variable name or column name for elevation differences depending if output
        is xarray.Dataset or GeoDataFrame. Default to 'elev_diff'

    Returns
    -------
    xr.Dataset | gpd.GeoDataFrame
        Source dataset with added elevation differences

    Notes
    -----
    This function does not reproject data (it assumes source and reference have the same CRS).
    """
    if isinstance(source, xr.Dataset):
        source_elev_diff = source.elevation - reference.elevation
        source[diff_col_name] = source_elev_diff

        return source

    if isinstance(source, gpd.GeoDataFrame):
        # point source
        if source_col is None:
            msg_val_error = "source_col must be specified when source is GeoDataFrame"
            raise ValueError(msg_val_error)
        # sample reference at point locations
        da_points = source.get_coordinates().to_xarray()
        samples = reference.interp(da_points).to_dataframe()
        source["elev_diff"] = (
            source[source_col].to_numpy() - samples["elevation"].to_numpy()
        )
        return source

    msg_type_error = "source must be xr.Dataset or gpd.GeoDataFrame"
    raise TypeError(msg_type_error)


def hillshade(
    da: xr.DataArray, azi: int = 45, alt: int = 45
) -> tuple[tuple[str, ...], np.ndarray]:
    """
    Create hillshade array from DEM.
    Pass with dem['hillshade'] = coincident.plot.utils.hillshade(dem.elevation)

    Parameters
    ----------
    da: xarray.DataArray
        Dataarray containing elevation values
    azi: int
        The shading azimuth in degrees (0-360°) going clockwise, starting from north.
    alt: int
        The shading altitude in degrees (0-90°). 90° is straight from above.

    Returns
    -------
    tuple
        (dims, data) tuple where dims are dimension names and data is uint8 array
    """
    # Ensure 2D array by squeezing extra dimensions
    da = da.squeeze()

    x, y = np.gradient(da)
    slope = np.pi / 2.0 - np.arctan(np.sqrt(x * x + y * y))
    aspect = np.arctan2(-x, y)
    azimuth = 360.0 - azi
    azimuthrad = azimuth * np.pi / 180.0
    altituderad = alt * np.pi / 180.0

    shaded = np.sin(altituderad) * np.sin(slope) + np.cos(altituderad) * np.cos(
        slope
    ) * np.cos(azimuthrad - aspect)

    data = 255 * (shaded + 1) / 2
    data[data < 0] = 0  # set hillshade values to min of 0.

    return da.dims, data.astype(np.uint8)



from __future__ import annotations

import importlib.metadata

project = "coincident"
copyright = "2025, UW TACO Lab"
author = "Scott Henderson"
version = release = importlib.metadata.version("coincident")

# NOTE: it seems order of extensions matters if one depends on another
extensions = [
    # "myst_parser",
    "myst_nb",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
    "sphinx_design",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".ipynb": "myst-nb",
    ".md": "myst-nb",
}
exclude_patterns = [
    "_build",
    "**.ipynb_checkpoints",
    "Thumbs.db",
    ".DS_Store",
    ".env",
    ".venv",
]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# Other API docs using this theme for inspiration:
# https://virtualizarr.readthedocs.io

html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "use_edit_page_button": True,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/uw-cryo/coincident",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        },
    ],
}
html_title = "Coincident"
html_context = {
    "github_user": "uw-cryo",
    "github_repo": "coincident",
    "github_version": "main",
    "doc_path": "docs",
}

# remove sidebar, see GH issue #82
html_css_files = [
    "custom.css",
]

html_static_path = ["_static"]

# NOTE: consider adding back in once for distinct sections (user guide, examples, API reference)
# https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/layout.html#primary-sidebar-left
# html_sidebars = {"**": []}
# Remove primary sidebar from just API autodoc page
html_sidebars = {"api": []}

myst_enable_extensions = [
    "colon_fence",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "geopandas": ("https://geopandas.org/en/latest", None),
    "pystac_client": ("https://pystac-client.readthedocs.io/en/stable", None),
    "pystac": ("https://pystac.readthedocs.io/en/stable", None),
    "stac-asset": ("https://stac-asset.readthedocs.io/en/latest", None),
    "xarray": ("https://docs.xarray.dev/en/stable", None),
}

nitpick_ignore = [
    ("py:class", "_io.StringIO"),
    ("py:class", "_io.BytesIO"),
]

always_document_param_types = True
# autodoc_typehints = "none"
nb_execution_mode = "auto"  # off, on
nb_execution_show_tb = True
nb_execution_timeout = 90
# nb_execution_excludepatterns = ["elevation_plotting.ipynb"]



from __future__ import annotations

import pytest
import xarray as xr

import coincident
from coincident.io.xarray import to_dataset

network = pytest.mark.network


@network
def test_to_dataset_with_cop30_lazy(aoi):
    """Test `to_dataset` functionality with COP30 dataset."""
    gf_cop30 = coincident.search.search(
        dataset="cop30",
        intersects=aoi,
    )
    ds = to_dataset(
        gf_cop30,
        aoi=aoi,
        resolution=0.1,  # ~1km
    )
    assert isinstance(ds, xr.Dataset), "Expected output to be an xarray Dataset."
    assert "data" in ds.data_vars, "Expected 'data' variable in the Dataset."


@network
def test_to_dataset_with_worldcover_lazy(aoi):
    """Test `to_dataset` functionality with WorldCover dataset."""
    gf_wc = coincident.search.search(
        dataset="worldcover",
        intersects=aoi,
        datetime=["2020"],
    )
    ds = to_dataset(
        gf_wc,
        bands=["map"],
        aoi=aoi,
        resolution=0.1,  # ~1km
    )
    assert isinstance(ds, xr.Dataset), "Expected output to be an xarray Dataset."
    assert "map" in ds.data_vars, "Expected 'map' variable in the Dataset."



"""
Searching for NCALM and NOAA data with an input search polygon is made easy via the OpenTopography API
https://portal.opentopography.org/apidocs/
There is also a GitHub repo containing flight extent polygons for NOAA
https://github.com/OpenTopography/Data_Catalog_Spatial_Boundaries/tree/main/NOAA_Coastal_Lidar
"""

from __future__ import annotations

from datetime import datetime

import geopandas as gpd
import pandas as pd
import requests  # type: ignore[import-untyped]
from pandas import Timestamp
from shapely.geometry import shape


def search_ncalm_noaa(
    aoi: gpd.GeoDataFrame | gpd.GeoSeries,
    search_start: Timestamp,
    search_end: Timestamp,
    dataset: str | None = None,
) -> gpd.GeoDataFrame:
    """
    Perform a search for NCALM LiDAR or NOAA Coastal LiDAR footprints and metadata via
    the OpenTopography API.

    Parameters
    ----------
    aoi : gpd.GeoDataFrame | gpd.GeoSeries
        A GeoDataFrame or GeoSeries containing a geometry to restrict the search area, by default does a global search.
    search_start : Timestamp, optional
        The start datetime for the search, by default searches the entire catalog defined by 'dataset'.
    search_end : Timestamp, optional
        The end datetime for the search, by default searches the entire catalog defined by 'dataset'.
    dataset : str
        The dataset type (either "noaa" or "ncalm").

    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame containing the search results.

    Raises
    ------
    ValueError
        If no valid `aoi` is provided or the API request fails.
    """
    # If `aoi` is None, set the entire world as the search area
    if aoi is None:
        globe = shape(
            {
                "type": "Polygon",
                "coordinates": [
                    [
                        [-180, -90],
                        [180, -90],
                        [180, 90],
                        [-180, 90],
                        [-180, -90],
                    ]
                ],
            }
        )
        search_poly = gpd.GeoSeries(globe, crs="EPSG:4326")
        coords = ",".join(
            [f"{x},{y}" for x, y in search_poly.geometry[0].exterior.coords]
        )
    else:
        # convex_hull works better than simplify for more-complex geometries (ie. Louisiana)
        # https://raw.githubusercontent.com/unitedstates/districts/refs/heads/gh-pages/states/LA/shape.geojson
        search_poly = aoi.to_crs(4326).union_all()
        search_poly_chull = search_poly.convex_hull
        coords = ",".join([f"{x},{y}" for x, y in search_poly_chull.exterior.coords])

    if dataset not in ["noaa", "ncalm"]:
        msg = f"Unsupported dataset: {dataset}"
        raise ValueError(msg)

    # https://requests.readthedocs.io/en/latest/user/quickstart/#passing-parameters-in-urls
    url_api_base = "https://portal.opentopography.org/API/otCatalog"
    params_api = {
        "productFormat": "PointCloud",
        "detail": "true",
        "outputFormat": "json",
        "polygon": coords,
        "include_federated": "true" if dataset == "noaa" else "false",
    }

    response = requests.get(url_api_base, params=params_api, timeout=30)
    if response.status_code != 200:
        msg_response = f"Error querying OpenTopography API: {response.status_code}"
        raise ValueError(msg_response)

    catalog = response.json()
    if dataset == "noaa":
        catalog = [
            ds
            for ds in catalog["Datasets"]
            if "NOAA" in ds["Dataset"]["identifier"]["propertyID"]
            and ds["Dataset"]["temporalCoverage"] != ""
        ]
    else:
        catalog = [
            ds for ds in catalog["Datasets"] if ds["Dataset"]["temporalCoverage"] != ""
        ]

    # temporal filtering
    start_dt, end_dt = search_start.to_pydatetime(), search_end.to_pydatetime()
    filtered_catalog = [
        entry
        for entry in catalog
        if any(
            start_dt <= datetime.fromisoformat(date) <= end_dt
            for date in (entry["Dataset"]["temporalCoverage"].split(" / "))
        )
    ]

    # convert the filtered catalog into a GeoDataFrame
    gf_lidar = gpd.GeoDataFrame(
        [
            {
                "id": entry["Dataset"]["identifier"]["value"],
                "title": entry["Dataset"]["name"],
                "start_datetime": datetime.fromisoformat(
                    entry["Dataset"]["temporalCoverage"].split(" / ")[0]
                ),
                "end_datetime": datetime.fromisoformat(
                    entry["Dataset"]["temporalCoverage"].split(" / ")[-1]
                ),
                "geometry": shape(
                    entry["Dataset"]["spatialCoverage"]["geo"]["geojson"]["features"][
                        0
                    ]["geometry"]
                ),
            }
            for entry in filtered_catalog
        ],
        crs="EPSG:4326",
    )

    # filter the GeoDataFrame to only include datasets that intersect with the search polygon
    return gf_lidar[gf_lidar.intersects(search_poly)].reset_index(drop=True)


# TODO: add meaningful error messages when search fails
def search_opentopo(
    dataset: str | None,
    intersects: gpd.GeoDataFrame | gpd.GeoSeries,
    search_start: Timestamp | None = None,
    search_end: Timestamp | None = None,
) -> gpd.GeoDataFrame:
    """
    Integrate the OpenTopography dataset search into the main search function.

    Parameters
    ----------
    dataset : str
        Dataset alias ('noaa' or 'ncalm').
    intersects : gpd.GeoDataFrame | gpd.GeoSeries
        The geometry to restrict the search.
    search_start : Timestamp, optional
        The start time for the temporal search, by default None.
    search_end : Timestamp, optional
        The end time for the temporal search, by default None.
    **kwargs : Any
        Additional parameters for the search.

    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame containing the search results from the OpenTopography API.
    """
    if search_start is None and search_end is None:
        search_start = pd.Timestamp(
            "1996-01-01"
        )  # Starting from January 1, 1996 (first NOAAA dataset)
        search_end = pd.Timestamp.today()  # Default to today's date

    # search using the OpenTopography API
    return search_ncalm_noaa(
        intersects, search_start=search_start, search_end=search_end, dataset=dataset
    )



# ruff: noqa: ARG001
from __future__ import annotations

import geopandas as gpd
import pytest
import xarray as xr

# import os
# if not os.environ.get('MAXAR_API_KEY'):
#    os.environ['MAXAR_API_KEY'] = 'fake-test-key'


@pytest.fixture(scope="package")
def aoi():
    # 11 vertices, 1,361km^2
    aoi_url = "https://raw.githubusercontent.com/SlideRuleEarth/sliderule-python/main/data/grandmesa.geojson"
    return gpd.read_file(aoi_url)


@pytest.fixture
def large_aoi(scope="package"):
    # 260 vertices, large area 269,590 km^2
    aoi_url = "https://raw.githubusercontent.com/unitedstates/districts/refs/heads/gh-pages/states/CO/shape.geojson"
    return gpd.read_file(aoi_url)


# TODO: add a small geojson AOI for testing bathy sites
@pytest.fixture
def bathy_aoi(scope="package"):
    # 631 vertices, large area ~ 27,000 km^2
    aoi_url = "https://raw.githubusercontent.com/unitedstates/districts/refs/heads/gh-pages/states/MA/shape.geojson"
    return gpd.read_file(aoi_url)


@pytest.fixture(scope="package")
def tinyaoi():
    # 4 vertices
    return gpd.read_file("tests/data/tiny.geojson")


@pytest.fixture(scope="package")
def dem_tiny():
    # from CO PCD site
    ds_cop30 = xr.open_dataset("tests/data/dem_tiny.tif")
    return ds_cop30.rename({"band_data": "elevation"})


@pytest.fixture(scope="package")
def points_tiny():
    # random points as is2/gedi facade overlapping and same CRS as dem_tiny
    # has cols h_li, elevation_lm, and elevation_hr
    # ^ random elevation values
    return gpd.read_file("tests/data/points_tiny.gpkg")


@pytest.fixture(scope="package")
def dem_tiny_utm():
    ds_cop30 = xr.open_dataset("tests/data/dem_tiny_utm.tif")
    return ds_cop30.rename({"band_data": "elevation"})


@pytest.fixture(scope="package")
def points_tiny_utm():
    return gpd.read_file("tests/data/points_tiny_utm.gpkg")



"""
STAC Search Functions
"""

from __future__ import annotations

import warnings
from typing import Any

import geopandas as gpd
import pystac
import pystac_client
import stac_geoparquet

# Any import that requires auth to work will be optional
try:
    import maxar_platform
except ImportError:
    msg_notfound = "'maxar-platform' package not found. Install for maxar functionality: https://pypi.org/project/maxar-platform/"
    warnings.warn(msg_notfound, stacklevel=2)

try:
    import maxar_platform.discovery
except maxar_platform.session.NoSessionCredentials:
    msg_noauth = "Unable to authenticate with Maxar API. Please set MAXAR_API_KEY environment variable."
    warnings.warn(msg_noauth, stacklevel=2)
except maxar_platform.exceptions.UnAuthorizedException:
    msg_noauth = (
        "Unable to authenticate with Maxar API. Please check MAXAR_API_KEY is valid."
    )
    warnings.warn(msg_noauth, stacklevel=2)


def _filter_assets(assets: gpd.GeoDataFrame) -> dict[str, str]:
    """Remove key:None pairs from assets"""
    keep_keys = []
    for k, v in assets.items():
        if v is not None:
            keep_keys.append(k)

    return {key: assets[key] for key in keep_keys}


def to_geopandas(
    collection: pystac.item_collection.ItemCollection,
) -> gpd.GeoDataFrame:
    """
    Convert a STAC ItemCollection to a GeoDataFrame.
    This function converts a given STAC ItemCollection to a GeoDataFrame using the
    `stac_geoparquet.arrow.parse_stac_items_to_arrow` method. It also adds an additional
    column 'dayofyear' for convenience.

    Parameters
    ----------
    collection : pystac.item_collection.ItemCollection
        The STAC ItemCollection to be converted.

    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame containing the data from the STAC ItemCollection.

    Raises
    ------
    ValueError
        If the provided ItemCollection is empty.
    """
    # Catch if no items are passed
    if len(collection) == 0:
        message = "ItemCollection is empty, cannot convert to GeoDataFrame"
        raise ValueError(message)

    record_batch_reader = stac_geoparquet.arrow.parse_stac_items_to_arrow(collection)
    gf = gpd.GeoDataFrame.from_arrow(record_batch_reader)  # doesn't keep arrow dtypes

    # Workaround stac-geoparquet limitation https://github.com/stac-utils/stac-geoparquet/issues/82
    gf["assets"] = gf["assets"].apply(_filter_assets)

    # Additional columns for convenience
    # NOTE: these become entries under STAC properties
    gf["dayofyear"] = gf["datetime"].dt.dayofyear

    return gf


def to_pystac_items(gf: gpd.GeoDataFrame) -> list[pystac.Item]:
    """
    Converts a :class:`~geopandas.GeoDataFrame` to a list of PySTAC Items.

    Parameters
    ----------
    gf : GeoDataFrame
        The GeoDataFrame to be converted.

    Returns
    -------
    list[pystac.Item]
        A list of PySTAC Items created from the GeoDataFrame.
    """

    batch = stac_geoparquet.arrow.stac_table_to_items(gf.to_arrow())

    return [pystac.Item.from_dict(x) for x in batch]


def search(
    client: pystac_client.client.Client, **kwargs: dict[str, Any] | None
) -> pystac_client.item_search.ItemSearch:
    """Search any STAC API (e.g. https://github.com/nasa/cmr-stac)"""
    # NOTE: add logging for kwargs?
    # print(kwargs)
    results = client.search(
        **kwargs,
    )
    return results.item_collection()  # actually run the query


def configure_maxar_client(area_based_calc: bool = True) -> pystac_client.client.Client:
    # automatically checks authentication

    client = maxar_platform.discovery.open_catalog(catalog="imagery")
    # Custom Maxar setting
    client.area_based_calc = area_based_calc

    return client


def configure_stac_client(url: str) -> pystac_client.client.Client:
    return pystac_client.Client.open(url=url)



from __future__ import annotations

import coincident as m


def test_version():
    # Just check that conventional __version__ exists
    assert m.__version__
    # NOTE: these do not necessarily match during development I think
    # importlib picks up the version of package at time of install
    # but _version.py / __init__.py is only updated by `pipx run build`
    # https://github.com/python/cpython/issues/94181
    # assert importlib.metadata.version("coincident") == m.__version__



from __future__ import annotations

import geopandas as gpd
import pytest

import coincident.overlaps


@pytest.fixture
def geodataframe():
    """1x1 degree ll corner at (0,0)"""
    fc = {
        "type": "FeatureCollection",
        "features": [
            {
                "id": "0",
                "properties": {
                    "start_datetime": gpd.pd.Timestamp("2021-01-01"),
                    "end_datetime": gpd.pd.Timestamp("2021-01-06"),
                },
                "type": "Feature",
                "geometry": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [0.5, -0.5],
                            [0.5, 0.5],
                            [-0.5, 0.5],
                            [-0.5, -0.5],
                            [0.5, -0.5],
                        ]
                    ],
                },
            }
        ],
    }
    return gpd.GeoDataFrame.from_features(fc, crs="EPSG:4326")


def test_geographic_area(geodataframe):
    area = (
        (coincident.overlaps.geographic_area(geodataframe) * 1e-6)
        .to_numpy()
        .astype(int)
    )
    assert area == 12309


def test_subset_by_minimum_area(geodataframe):
    big_enough = coincident.overlaps.subset_by_minimum_area(geodataframe)
    too_small = coincident.overlaps.subset_by_minimum_area(geodataframe, 20100)
    assert len(big_enough) == 1
    assert len(too_small) == 0


def test_subset_by_temporal_overlap(geodataframe):
    start = gpd.pd.Timestamp("2020-12-01")
    end = gpd.pd.Timestamp("2020-12-31")
    nonoverlapping = coincident.overlaps.subset_by_temporal_overlap(
        geodataframe, start_datetime=start, end_datetime=end, temporal_buffer=0
    )
    overlapping = coincident.overlaps.subset_by_temporal_overlap(
        geodataframe, start_datetime=start, end_datetime=end, temporal_buffer=5
    )
    assert len(nonoverlapping) == 0
    assert len(overlapping) == 1



"""
USGS 3DEP LiDAR

https://www.usgs.gov/3d-elevation-program
"""

# from pydantic.dataclasses import dataclass
from __future__ import annotations

from dataclasses import dataclass

from coincident.datasets.general import Dataset


@dataclass
class ThreeDEP(Dataset):
    """Essential metadata for USGS 3DEP"""

    has_stac_api: bool = False
    search: str = (
        # "https://prd-tnm.s3.amazonaws.com/StagedProducts/Elevation/metadata/WESM.gpkg"
        "s3://prd-tnm/StagedProducts/Elevation/metadata/WESM.gpkg"
    )
    start: str = "2000-12-01"
    end: str | None = None
    type: str = "lidar"
    alias: str = "3dep"
    provider: str = "usgs"



"""Helper functions for plot testing"""

from __future__ import annotations

import matplotlib.patches
import matplotlib.pyplot as plt
import pytest


@pytest.fixture
def assert_boxplot():
    """Fixture for common boxplot assertions"""

    def _assert_boxplot(axd, expected_boxes):
        """Helper to verify boxplot properties"""
        assert isinstance(
            axd, plt.Axes
        ), "Return value should be a matplotlib Axes object"
        assert len(axd.get_children()) > 0, "Figure should exist"
        assert (
            len(
                [
                    child
                    for child in axd.get_children()
                    if isinstance(child, matplotlib.patches.PathPatch)
                ]
            )
            == expected_boxes
        ), f"Should have {expected_boxes} boxplot boxes"
        assert axd.get_legend() is not None, "Legend should exist"
        assert len(axd.figure.axes) == 2, "Should have two y-axes"
        assert (
            axd.figure.axes[1].get_ylabel() == "Count"
        ), "Second y-axis should be 'Count'"

    return _assert_boxplot



"""
coincident: Search and analysis of STV Precursor Coincident Datasets
"""

from __future__ import annotations

from coincident.io import download, sliderule, xarray

__all__ = ["sliderule", "xarray", "download"]



from __future__ import annotations

import asyncio
from pathlib import Path

import pytest

import coincident

network = pytest.mark.network
try:
    import maxar_platform.discovery  # noqa: F401

    not_authenticated = False
except:  # noqa: E722
    not_authenticated = True
maxar_authenticated = pytest.mark.skipif(
    not_authenticated, reason="Not authenticated with Maxar API"
)


@network
@maxar_authenticated
@pytest.mark.filterwarnings("ignore:the actual content type does not match")
@pytest.mark.skip(reason="temporarily skip to see if CI failing due to async code")
def test_download_maxar_browse():
    gf = coincident.search.search(dataset="maxar", ids=["102001008EC5AC00"])
    item = coincident.search.stac.to_pystac_items(gf)[0]
    asyncio.run(coincident.io.download.download_item(item, config="maxar"))
    assert Path("/tmp/102001008EC5AC00.browse.tif").exists()
    assert Path("/tmp/102001008EC5AC00.json").exists()


# @network
# def test_download_planetary_computer():
#     print('todo: small cop30 asset?')



"""
Microsoft Planetary Computer

https://planetarycomputer.microsoft.com/docs/quickstarts/reading-stac/

- https://planetarycomputer.microsoft.com/dataset/esa-worldcover
- https://planetarycomputer.microsoft.com/dataset/cop-dem-glo-30
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from coincident.datasets.general import Dataset

STACAPI = "https://planetarycomputer.microsoft.com/api/stac/v1"


@dataclass
class COP30(Dataset):
    """Essential metadata and data access for Copernicus DEM"""

    alias: str = "cop30"
    has_stac_api: bool = True
    collections: list[str] = field(default_factory=lambda: ["cop-dem-glo-30"])
    search: str = STACAPI
    start: str | None = None  # Copernicus DEM has 'representative' datetime: 2021-04-22
    end: str | None = None
    type: str = "dem"
    provider: str = "microsoft"


@dataclass
class WorldCover(Dataset):
    """Essential metadata for ESA WorldCover"""

    alias: str = "worldcover"
    has_stac_api: bool = True
    collections: list[str] = field(default_factory=lambda: ["esa-worldcover"])
    search: str = STACAPI
    start: str = "2020-01-01"
    end: str = "2021-12-31"
    type: str = "lulc"
    provider: str = "microsoft"
    classmap: dict[int, Any] = field(
        default_factory=lambda: {
            10: {"hex": "#006400", "description": "Tree cover"},
            20: {"hex": "#FFBB22", "description": "Shrubland"},
            30: {"hex": "#FFFF4C", "description": "Grassland"},
            40: {"hex": "#F096FF", "description": "Cropland"},
            50: {"hex": "#FA0000", "description": "Built-up"},
            60: {"hex": "#B4B4B4", "description": "Bare / sparse vegetation"},
            70: {"hex": "#F0F0F0", "description": "Snow and ice"},
            80: {"hex": "#0064C8", "description": "Permanent water bodies"},
            90: {"hex": "#0096A0", "description": "Herbaceous wetland"},
            95: {"hex": "#00CF75", "description": "Mangroves"},
            100: {"hex": "#FAE6A0", "description": "Moss and lichen"},
        }
    )



from __future__ import annotations

import geopandas as gpd
import pytest

import coincident

# import sliderule
# sliderule.init(verbose=True)

network = pytest.mark.network


def test_gdf_to_sliderule_polygon():
    gf = gpd.GeoDataFrame(
        geometry=gpd.points_from_xy(
            [-106.46953038438801, -106.47209604357627],
            [38.781089777057986, 38.78292421778072],
            crs="EPSG:4326",
        )
    )
    poly = coincident.io.sliderule._gdf_to_sliderule_polygon(gf)
    assert isinstance(poly, list)
    assert len(poly) == 5


@network
@pytest.mark.skip(reason="https://github.com/uw-cryo/coincident/issues/35")
class TestSlideRule:
    def test_subset_gedi02a(self, tinyaoi):
        gf_gedi = coincident.search.search(
            dataset="gedi", intersects=tinyaoi, datetime="2022"
        )
        data = coincident.io.sliderule.subset_gedi02a(
            gf_gedi, aoi=tinyaoi, include_worldcover=True
        )
        assert isinstance(data, gpd.GeoDataFrame)
        assert "elevation_lm" in data.columns
        assert "elevation_hr" in data.columns
        assert data.iloc[0].name.toordinal() == 738192
        assert data.iloc[0]["worldcover.value"] == "Tree cover"
        assert data.shape == (262, 11)

    def test_subset_atl06(self, tinyaoi):
        gf_is2 = coincident.search.search(
            dataset="icesat-2", intersects=tinyaoi, datetime="2022"
        )
        data = coincident.io.sliderule.subset_atl06(gf_is2, aoi=tinyaoi)
        assert isinstance(data, gpd.GeoDataFrame)
        assert "h_li" in data.columns
        assert 1034 in data.rgt.unique()
        assert data.iloc[0].name.toordinal() == 738304
        assert data.shape == (413, 21)

    def test_process_atl06sr(self, tinyaoi):
        gf_is2 = coincident.search.search(
            dataset="icesat-2", intersects=tinyaoi, datetime="2022"
        )
        data = coincident.io.sliderule.process_atl06sr(
            gf_is2, aoi=tinyaoi, include_3dep=True
        )
        assert isinstance(data, gpd.GeoDataFrame)
        assert "n_fit_photons" in data.columns
        assert "3dep.value" in data.columns
        assert data.shape == (367, 20)
        assert data.iloc[0].name.toordinal() == 738304
        assert int(data.iloc[0].h_mean) == 3178

    @pytest.mark.filterwarnings(
        "ignore:Geometry is in a geographic CRS:UserWarning:geopandas"
    )
    def test_sample_3dep(self):
        gf = gpd.GeoDataFrame(
            geometry=gpd.points_from_xy(
                [-106.46953038438801, -106.47209604357627],
                [38.781089777057986, 38.78292421778072],
                crs="EPSG:4326",
            )
        )
        samples = coincident.io.sliderule.sample_3dep(gf)
        assert {"file", "geometry", "time", "value"} == set(samples.columns)
        assert samples.time.dtype == "<M8[ns]"
        assert str(samples.time.dt.date.iloc[0]) == "2019-09-04"
        # NOTE: is this on edge of 2 tiles?! sometimes comes back as {3110, 3178}...
        # assert set(samples.value.astype('i2')) == {3111, 3178}



"""
NASA Commercial SmallSat Data Acquisition Program (CSDA)

https://www.earthdata.nasa.gov/esds/csda

TSX/TDX = https://csdap.earthdata.nasa.gov/stac/collections/airbus
# TODO: compare this archive against
# https://api.oneatlas.airbus.com/api-catalog-v2/radar/overview/index.html

Notes
https://github.com/stac-extensions/sar?tab=readme-ov-file#product-type
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from coincident.datasets.general import Dataset


@dataclass
class TDX(Dataset):
    """Essential metadata for TanDEM-X / TerraSAR-X
    platform =
     - PAZ (PAX)
     - TerraSAR-X (TSX-1)
     - TanDEM-X (TDX-1)

    sar:product_type =
    - Coregistered Single Look Slant Range Complex (COSSC)
    - Geocoded Ellipsoid Corrected (GEC)
    - Single Look Slant Range Complex (SSC)
    - Enhanced Ellipsoid Corrected (EEC)

    sar:instrument_mode =
     - Staring Spotlight (ST)
     - High Resolution Spotlight (HS)
     - Spotlight (SL)
     - Stripmap (SM)
     - ScanSAR (SC)
     - Wide ScanSAR (WSC)

    Other properties to filter on:
    sar:polarizations = ['HH']
    orbitDirection = ASCENDING / DESCENDING
    """

    alias: str = "tdx"
    has_stac_api: bool = True
    collections: list[str] = field(default_factory=lambda: ["airbus"])
    search: str = "https://csdap.earthdata.nasa.gov/stac"
    start: str = "2007-07-01"  # TODO: check this
    end: str | None = None
    type: str = "sar"
    provider: str = "csda"
    stac_kwargs: dict[str, Any] = field(
        default_factory=lambda: {
            "limit": 1000,
            "filter": {
                "op": "and",
                "args": [
                    # exclude PAZ, only SSC products
                    {
                        "op": "in",
                        "args": [{"property": "platform"}, ["TDX-1", "TSX-1"]],
                    },
                    {"op": "=", "args": [{"property": "sar:product_type"}, "SSC"]},
                ],
            },
        }
    )



"""
Maxar VHR stereo imagery

https://www.maxar.com/maxar-intelligence/products/satellite-imagery
"""

# from pydantic.dataclasses import dataclass, Field # type: ignore[attr-defined]
from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum

from coincident.datasets.general import Dataset


class Collection(str, Enum):
    """wv01,wv02,wv03-vnir,ge01"""

    wv01 = "wv01"
    wv02 = "wv02"
    wv03_vnir = "wv03-vnir"
    ge01 = "ge01"


@dataclass
class Stereo(Dataset):
    """Essential metadata for Maxar In Track Stereo"""

    alias: str = "maxar"
    has_stac_api: bool = True
    collections: list[Collection] = field(
        default_factory=lambda: ["wv01", "wv02", "wv03-vnir", "ge01"]
    )  # type: ignore[assignment]
    search: str = "https://api.maxar.com/discovery/v1/search"
    start: str = "2007-01-01"
    end: str | None = None
    type: str = "stereo"
    # Unique to Maxar
    area_based_calc: bool = False
    provider: str = "maxar"



"""
Searching for NEON data with spatial and temporal filters is actually quite difficult.
I couldn't find any STAC catalogs or code to make this easier

https://data.neonscience.org/data-products/DP3.30024.001
https://www.neonscience.org/resources/learning-hub/tutorials/neon-api-intro-requests-py

The process goes as follows:
- Use requests to get some basic metadata for all NEON sites that contain discrete return LiDAR point cloud data
    - Here, we grab the site code, site name, available months, and product URLs
    - Note that grabbing the specific collection time for each flight requires
        a lot more key-value pair digging and i do not believe is worth the effort
    - We grab the 'centroid' (top left of the center-most tile of footprint) here as well
        this is because there is no easy way to search on an input geometry and we can
        use the 'centroid' to make that process easier
- Subset the resulting gdf on the user-input start and end times
- Subset that gdf by only including sites that have 'centroids' within 1 degree (~111km)
    we do this because grabbing the actual footprint geometry is time-consuming.
    the only way to do this dynamically, at least barring the use of
    more unnecessary and unique llibraries/packages from what i found,
    is by making a GET request for each site and then reading in the footprint
    hosted on storage.googleapis.com
- Grab the footprint geometry for each site
    These are hosted on storage.googleapis.com in the form of a shapefile
    that has all of the complex tile footprints as their own row geometries
"""

from __future__ import annotations

import geopandas as gpd
import pandas as pd
import pyogrio
import requests  # type: ignore[import-untyped]
from pandas import Timestamp
from shapely.geometry import Point


def build_neon_point_gf(sites_url: str) -> gpd.GeoDataFrame:
    """
    Build a GeoDataFrame containing NEON site information, including product URLs and available months.

    Parameters
    ----------
    sites_url : str
        The URL to the NEON API for retrieving site data.

    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame with site information and 'centroid' geometries. Note that these centroids
        are the top left of the center-most tile for each NEON flight. Also note that there is no
        easy way to pull the flight footprint geometry from NEON.
    """
    sites_request = requests.get(sites_url, timeout=30)
    sites_json = sites_request.json()

    df_neon = pd.concat(
        [
            pd.DataFrame(
                {
                    "id": site["siteCode"],
                    "title": site["siteName"],
                    "start_datetime": dict_lidar["availableMonths"],
                    "end_datetime": dict_lidar["availableMonths"],
                    "product_url": dict_lidar["availableDataUrls"],
                    "latitude": site["siteLatitude"],
                    "longitude": site["siteLongitude"],
                }
            )
            for site in sites_json["data"]
            for dict_lidar in [
                next(
                    (
                        item
                        for item in site["dataProducts"]
                        if item["dataProductCode"] == "DP3.30024.001"
                    ),
                    None,
                )
            ]
            if dict_lidar is not None
        ],
        ignore_index=True,
    )

    df_neon["geometry"] = df_neon.apply(
        lambda row: Point(row["longitude"], row["latitude"]), axis=1
    )
    gf_neon = gpd.GeoDataFrame(df_neon, geometry="geometry", crs="EPSG:4326")

    return gf_neon.drop(columns=["latitude", "longitude"])


def temporal_filter_neon(
    gf_neon: gpd.GeoDataFrame, search_start: Timestamp, search_end: Timestamp
) -> gpd.GeoDataFrame:
    """
    Filter NEON GeoDataFrame by temporal range.

    Parameters
    ----------
    gf_neon : gpd.GeoDataFrame
        The GeoDataFrame containing NEON site data from build_neon_point_gf().
    search_start : pd.Timestamp
        The start of the time range to filter by.
    search_end : pd.Timestamp
        The end of the time range to filter by.

    Returns
    -------
    gpd.GeoDataFrame
        A filtered GeoDataFrame based on the temporal range.
    """
    return gf_neon[
        (
            gf_neon["start_datetime"].apply(lambda x: pd.to_datetime(x + "-01"))
            >= search_start
        )
        & (
            gf_neon["start_datetime"].apply(lambda x: pd.to_datetime(x + "-01"))
            <= search_end
        )
    ].reset_index(drop=True)


def get_neon_bboxes(url: str, fallback_geometry: gpd.GeoSeries) -> gpd.GeoSeries:
    """
    Fetch and return bounding boxes for NEON data products.

    Parameters
    ----------
    url : str
        The URL to the NEON product data.
    fallback_geometry : gpd.GeoSeries
        A fallback geometry to return if footprint cannot be retrieved.

    Returns
    -------
    gpd.GeoSeries
        The bounding box geometry for the NEON data product.
    """
    response = requests.get(url, timeout=30)
    if response.status_code != 200:
        return fallback_geometry
    data = response.json().get("data", {})
    files = data.get("files", [])
    # Find the file that ends with 'merged_tiles.shp' dynamically
    # I really can't think of a better, more-efficient way to do this after dissecting the NEON API
    # The merged tiles shapefiles have similar filepaths, but we're missing some extra metadata
    # that contains another ID needed to hardcode those paths effectively (see 'D02' in commented link below)
    shp_file_url = next(
        (f["url"] for f in files if f["name"].endswith("merged_tiles.shp")), None
    )

    try:
        df_tiles = pyogrio.read_dataframe(shp_file_url).to_crs(4326)
    except:  # noqa: E722
        # TODO: fix provisional footprint grab
        # the provisional (usually synonymous with new) NEON flights don't typically have merged_tiles.shp
        # i tried to iterate over all the different tiles and grab bboxes but too inefficient
        # we'll just return the point geometry for now
        return fallback_geometry

    try:
        return df_tiles.union_all().envelope
    except:  # noqa: E722
        # see below link for example where make_valid() is necessary
        # https://storage.googleapis.com/neon-aop-products/2019/FullSite/D02/2019_SCBI_3/Metadata/DiscreteLidar/TileBoundary/shps/2019_SCBI_3_merged_tiles.shp
        return df_tiles.geometry.make_valid().union_all().envelope


def search_bboxes(
    intersects: gpd.GeoDataFrame | gpd.GeoSeries,
    search_start: Timestamp,
    search_end: Timestamp,
) -> gpd.GeoDataFrame:
    """
    Perform a search for NEON LiDAR metadata and respective bbox footprints. Note that this search
    will take a while if you denote a large aoi or a large time range.

    Parameters
    ----------
    intersects : gpd.GeoDataFrame | gpd.GeoSeries
        The geometry to restrict the search. By default does a global search.
    search_start : pd.Timestamp
        The start of the time range to filter by. By default searches all dates in the catalog (don't do this)
    search_end : pd.Timestamp
        The end of the time range to filter by. By default searches all dates in the catalog (don't do this)

    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame with NEON metadata and respective bbox footprints
    """
    if search_start is None and search_end is None:
        search_start = pd.Timestamp(
            "2013-06-01"
        )  # Starting from June, 2005 (first NEON dataset)
        search_end = pd.Timestamp.today()  # Default to today's date
    # note that the above will result in the search taking a very long time
    # if you're searching a larger area
    if intersects is None:
        # Define global bounding box
        intersects = gpd.GeoSeries([Point(-180, -90).buffer(360)])

    sites_url = "http://data.neonscience.org/api/v0/sites"
    # Build NEON point GeoDataFrame
    gf_neon = build_neon_point_gf(sites_url)

    # Apply temporal filter
    gf_neon = temporal_filter_neon(gf_neon, search_start, search_end)

    # Find nearest footprint 'centroids' within the intersection geometry
    gf_neon = (
        gpd.sjoin_nearest(gf_neon, intersects, max_distance=1)
        .drop(columns=["index_right"])
        .reset_index(drop=True)
    )

    # Apply bounding box function
    gf_neon["geometry"] = gf_neon.apply(
        lambda row: get_neon_bboxes(row["product_url"], row["geometry"]), axis=1
    )

    # sjoin to make sure geometries actually overlap with input aoi
    return (
        gf_neon.sjoin(intersects).drop(columns=["index_right"]).reset_index(drop=True)
    )



"""
NEON Discrete LiDAR Datasets

https://data.neonscience.org/data-products/DP3.30024.001
"""

from __future__ import annotations

from dataclasses import dataclass

from coincident.datasets.general import Dataset


@dataclass
class NEON(Dataset):
    """Essential metadata for NEON LiDAR"""

    has_stac_api: bool = False
    search: str = "http://data.neonscience.org/api/v0/sites"
    start: str = "2005-06-01"
    end: str | None = None
    type: str = "lidar"
    alias: str = "neon"
    provider: str = "neon"



from __future__ import annotations

from collections.abc import Callable
from functools import wraps
from importlib.util import find_spec

# NOTE: not really sure about type hinting here!
# https://stackoverflow.com/questions/65621789/mypy-untyped-decorator-makes-function-my-method-untyped
from typing import Any, TypeVar, cast

F = TypeVar("F", bound=Callable[..., Any])


def depends_on_optional(module_name: str) -> Callable[[F], F]:
    """borrowed from https://stackoverflow.com/questions/27361427/how-to-properly-deal-with-optional-features-in-python
    Usage: add decorator above function that uses optional dependency
    Example: @depends_on_optional("matplotlib")
    """

    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(*args, **kwargs):  # type: ignore[no-untyped-def]
            spec = find_spec(module_name)
            if spec is None:
                message = (
                    f"Optional dependency {module_name} not found ({func.__name__})."
                )
                raise ImportError(message)
            return func(*args, **kwargs)

        return cast(F, wrapper)

    return decorator



"""
selection of useful plots with matplotlib
"""

from __future__ import annotations

from coincident.plot.matplotlib import (
    boxplot_aspect,
    boxplot_elevation,
    boxplot_slope,
    boxplot_terrain_diff,
    compare_dems,
    hist_esa,
    plot_altimeter_points,
    plot_dem,
    plot_diff_hist,
    plot_esa_worldcover,
    plot_maxar_browse,
)
from coincident.plot.utils import (
    get_elev_diff,
    hillshade,
)

__all__ = [
    "plot_esa_worldcover",
    "plot_maxar_browse",
    "hillshade",
    "plot_dem",
    "plot_altimeter_points",
    "get_elev_diff",
    "plot_diff_hist",
    "compare_dems",
    "boxplot_terrain_diff",
    "boxplot_slope",
    "boxplot_elevation",
    "boxplot_aspect",
    "hist_esa",
]



from __future__ import annotations

import coincident


def test_valid_data_sources():
    supported = {"maxar", "nasa", "usgs"}
    imported = set(dir(coincident.datasets))
    assert imported & supported == supported


def test_maxar_defaults():
    ds = coincident.datasets.maxar.Stereo()
    mono_collections = ["wv01", "wv02", "wv03-vnir", "ge01"]
    assert set(ds.collections) == set(mono_collections)


def test_threedep_defaults():
    ds = coincident.datasets.usgs.ThreeDEP()
    assert ds.alias == "3dep"


def test_opentopo_defaults():
    ds_noaa = coincident.datasets.opentopo.NOAA()
    assert ds_noaa.alias == "noaa"
    ds_ncalm = coincident.datasets.opentopo.NCALM()
    assert ds_ncalm.alias == "ncalm"


def test_neon_defaults():
    ds = coincident.datasets.neon.NEON()
    assert ds.alias == "neon"



"""
Convenience functions for downloading STAC items via stac-asset
"""

from __future__ import annotations

import os
from pathlib import Path

import stac_asset
from pystac import Item


async def download_item(
    item: Item,
    path: str = "/tmp",
    config: str | None = None,
) -> Item:
    """
    Downloads a STAC item to a specified local path.

    Parameters
    ----------
    item : pystac.Item
        The STAC item to be downloaded.
    path : str, optional
        The local directory path where the item will be downloaded. Default is "/tmp".
    config : str, optional
        If config=='maxar', a MAXAR-API-KEY HTTP Header is used for authentication.

    Returns
    -------
    pystac.Item
        The downloaded STAC item.

    Examples
    --------
    >>> localitem = asyncio.run(download_item(item, config=MAXAR_CONFIG))
    """
    posixpath = Path(path)

    if config == "maxar":
        config = stac_asset.Config(
            http_headers={"MAXAR-API-KEY": os.environ.get("MAXAR_API_KEY")}
        )
    # NOTE: prevent in-place modification of remote hrefs with item.clone()
    await stac_asset.download_item(item.clone(), posixpath, config=config)



"""
Selection of visualizations using matplotlib.
"""

from __future__ import annotations

import warnings
from typing import Any

import geopandas as gpd
import matplotlib.figure
import numpy as np
import pandas as pd
import pystac
import xarray as xr

# from matplotlib_scalebar.scalebar import ScaleBar
from scipy import stats
from shapely.geometry import box

try:
    import matplotlib.colors
    import matplotlib.pyplot as plt
    from matplotlib import cm
except ImportError:
    warnings.warn(
        "'matplotlib' package not found. Install for plotting functions: https://matplotlib.org/stable/install/index.html",
        stacklevel=2,
    )

from coincident._utils import depends_on_optional
from coincident.datasets.planetary_computer import WorldCover
from coincident.io.xarray import open_maxar_browse, to_dataset
from coincident.plot.utils import get_elev_diff
from coincident.search import search


@depends_on_optional("matplotlib")
def plot_esa_worldcover(
    ds: xr.Dataset,
    ax: plt.Axes | None = None,
    cax: plt.Axes | None = None,
    add_colorbar: bool = True,
    add_labels: bool = True,
) -> plt.Axes:
    """
    Map view of ESA WorldCover data.

    This function visualizes the ESA WorldCover dataset by mapping the class values to their recommended colors
    and descriptions.

    Parameters
    ----------
    ds : xr.Dataset
        An xarray Dataset containing the ESA WorldCover data.
    ax : plt.Axes
        The Matplotlib Axes on which the plot will be drawn.
    cax : plt.Axes
        The Matplotlib Axes used for the colorbar.
    add_colorbar : bool, optional
        Whether to include a colorbar in the plot, by default True.
    add_labels : bool, optional
        Whether to add labels to the plot, by default True.

    Returns
    -------
    plt.Axes
        The Matplotlib Axes object with the plot.

    Notes
    -----
    https://planetarycomputer.microsoft.com/dataset/esa-worldcover#Example-Notebook
    """
    # Custom categorical colormap for ESA WorldCover
    classmap = WorldCover().classmap

    colors = ["#000000" for r in range(256)]
    for key, value in classmap.items():
        colors[int(key)] = value["hex"]
    cmap = matplotlib.colors.ListedColormap(colors)
    values = list(classmap)
    boundaries = [(values[i + 1] + values[i]) / 2 for i in range(len(values) - 1)]
    boundaries = [0, *boundaries, 255]
    ticks = [
        (boundaries[i + 1] + boundaries[i]) / 2 for i in range(len(boundaries) - 1)
    ]
    tick_labels = [value["description"] for value in classmap.values()]
    normalizer = matplotlib.colors.Normalize(vmin=0, vmax=255)

    da = ds.to_dataarray().squeeze()

    if ax is None:
        fig, ax = plt.subplots()
        # set aspect ratio according to mid scene latitude
        if da.rio.crs.to_epsg() == 4326:
            mid_lat = da.latitude[int(da.latitude.size / 2)].to_numpy()  # PD011
            ax.set_aspect(aspect=1 / np.cos(np.deg2rad(mid_lat)))
    else:
        fig = ax.get_figure()

    da.plot(
        ax=ax, cmap=cmap, norm=normalizer, add_labels=add_labels, add_colorbar=False
    )

    if add_colorbar:
        # Specific to panel_plots, so rather than cax, add_to_panel=True boolean?
        if cax is not None:
            # Override standard xarray colorbar with custom one
            colorbar = fig.colorbar(
                cm.ScalarMappable(norm=normalizer, cmap=cmap),
                boundaries=boundaries,
                values=values,
                ax=cax,
                orientation="horizontal",
                pad=-1.1,  # hack to get colorbar in subplot mosaic in the right place
            )
            colorbar.set_ticks(ticks, labels=tick_labels, rotation=90)
        else:
            colorbar = fig.colorbar(
                cm.ScalarMappable(norm=normalizer, cmap=cmap),
                boundaries=boundaries,
                values=values,
                ax=ax,
            )
            colorbar.set_ticks(ticks, labels=tick_labels)

    return ax


@depends_on_optional("matplotlib")
def plot_maxar_browse(
    item: pystac.Item, ax: plt.Axes | None = None, overview_level: int = 0
) -> plt.Axes:
    """
    Map view of Maxar browse image from a STAC item using Matplotlib.

    Parameters
    ----------
    item : pystac.Item
        The STAC item containing the browse image to be plotted.
    overview_level : int, optional
        The overview level of the browse image to be opened, by default 0.

    Returns
    -------
    plt.Axes
        The Matplotlib Axes object with the plot.
    """

    da = open_maxar_browse(item, overview_level=overview_level)
    mid_lat = da.y[int(da.y.size / 2)].to_numpy()  # PD011

    if ax is None:
        _, ax = plt.subplots(figsize=(8, 11))
    da.plot.imshow(rgb="band", add_labels=False, ax=ax)
    ax.set_aspect(aspect=1 / np.cos(np.deg2rad(mid_lat)))
    ax.set_title(item.id)

    return ax


@depends_on_optional("matplotlib")
def _clear_labels(ax: plt.Axes) -> None:
    """
    Remove x and y axis labels.

    Parameters
    ----------
    ax : matplotlib.axes._axes.Axes
        The matplotlib axis object to remove labels from

    Returns
    -------
    None
        This function modifies the input axis object directly
    """
    ax.xaxis.set_major_formatter(plt.NullFormatter())
    ax.yaxis.set_major_formatter(plt.NullFormatter())
    ax.set_xlabel("")
    ax.set_ylabel("")


@depends_on_optional("matplotlib")
def plot_dem(
    da: xr.DataArray,
    ax: plt.Axes | None = None,
    cmap: str = "inferno",
    title: str = "",
    da_hillshade: xr.DataArray | None = None,
    **kwargs: Any,
) -> plt.Axes:
    """
    Map view of DEM with an option to plot a hillshade underneath

    Parameters
    ----------
    da: xarray.core.dataarray.DataArray
        A 2d xarray DataArray containing elevation data
    ax: matplotlib.axes._axes.Axes
        Axis on which to plot the DEM
    cmap: str
        Matplotlib colormap to use for the DEM
    title: str
        Title of the plot
    da_hillshade: xarray.core.dataarray.DataArray | None
        A 2d xarray DataArray containing hillshade data
    cmin: float | None
        Minimum value for the elevation colorbar
    cmax: float | None
        Maximum value for the elevation colorbar
    **kwargs: Any
        Artist properties passed to matplotlib.axes.Axes.imshow (e.g., vmin, vmax, etc.)

    Returns
    -------
    plt.Axes
        The Axes object with the DEM plot

    Notes
    -----
    alpha=0.5 set by default if you pass in a hillshade.
    """
    if ax is None:
        _, ax = plt.subplots()

    # hillshade with alpha=1.0
    if isinstance(da_hillshade, xr.DataArray):
        da_hillshade.plot.imshow(
            ax=ax, cmap="gray", alpha=1.0, add_colorbar=False, add_labels=False
        )
        if "alpha" not in kwargs:
            kwargs["alpha"] = 0.5
    # plot the DEM with alpha=0.5
    da.squeeze().plot.imshow(ax=ax, cmap=cmap, **kwargs)
    ax.set_title(title)

    return ax


@depends_on_optional("matplotlib")
def plot_altimeter_points(
    gf: gpd.GeoDataFrame,
    column: str,
    ax: plt.Axes | None = None,
    cmap: str = "inferno",
    title: str = "",
    da_hillshade: xr.DataArray | None = None,
    **kwargs: Any,
) -> plt.Axes:
    """
    Map view of laser altimeter point data with an optional hillshade background.

    Parameters
    ----------
    gf: geopandas.GeoDataFrame
        GeoDataFrame containing altimeter point data
    ax: matplotlib.axes._axes.Axes
        Axis on which to plot the points
    column: str
        Name of the GeoDataFrame elevation column to plot (e.g. 'h_li' for ICESat-2 ATL06)
    cmap: str
        Matplotlib colormap to use for the points
    alpha_p: float
        the alpha of the points
    title: str
        Title of the plot
    da_hillshade: xarray.core.dataarray.DataArray | None
        A 2d xarray DataArray containing hillshade data
    cmin: float | None
        Minimum value for the elevation colorbar
    cmax: float | None
        Maximum value for the elevation colorbar
    p_alpha: float | None
        Alpha value for the points, default is 0.5
    **kwargs: Any
        Artist properties passed to geopandas.GeoDataFrame.plot (e.g., markersize, legend, etc.)

    Returns
    -------
    plt.Axes
        The Axes object with the altimeter points plot

    Notes
    -----
    alpha=0.5 set by default if you pass in a hillshade.
    """
    if ax is None:
        _, ax = plt.subplots()

    # Plot the hillshade if available
    if isinstance(da_hillshade, xr.DataArray):
        da_hillshade.plot.imshow(
            ax=ax, cmap="gray", alpha=1.0, add_colorbar=False, add_labels=False
        )
        if "alpha" not in kwargs:
            kwargs["alpha"] = 0.5
    gf.plot(ax=ax, column=column, cmap=cmap, **kwargs)
    ax.set_title(title)
    return ax


def plot_diff_hist(
    da: xr.DataArray | pd.Series,
    ax: plt.Axes | None = None,
    dmin: float | None = -15,
    dmax: float | None = 15,
    **kwargs: Any,
) -> plt.Axes:
    """
    Histogram of elevation differences with median and mean lines.

    Parameters
    ----------
    da: xarray.DataArray | pd.Series
        DataArray or series containing elevation differences to plot
    ax: plt.Axes
        Axis on which to plot the histogram
    dmin: float
        Minimum value for histogram x-axis range, default -15
    dmax: float
        Maximum value for histogram x-axis range, default 15
    **kwargs: Any
        Artist properties passed to ax.hist()

    Returns
    -------
    plt.Axes
        The Axes object with the histogram plot
    """
    if ax is None:
        _, ax = plt.subplots()

    if isinstance(da, xr.DataArray):
        _ = da.plot.hist(ax=ax, bins=100, range=(dmin, dmax), color="gray", **kwargs)
    else:
        da.hist(ax=ax, bins=100, range=(dmin, dmax), color="gray", **kwargs)
    med = np.nanmedian(da)
    mean = np.nanmean(da)
    ax.axvline(0, color="k", lw=1)
    # NOTE: hardcoded unit label of meters
    ax.axvline(med, label=f"med: {med:.2f}m", color="cyan", lw=0.5)
    ax.axvline(mean, label=f"mean: {mean:.2f}m", color="magenta", lw=0.5)
    ax.set_xlabel("Elevation Difference (m)")
    ax.set_ylabel("Frequency")
    ax.legend(loc="best", fontsize="small")
    ax.grid(False)
    return ax


# TODO: Add add_hillshade=False?
@depends_on_optional("matplotlib")
def compare_dems(
    dem_list: list[xr.Dataset],
    gdf_dict: dict[str, tuple[gpd.GeoDataFrame, str]] | None = None,
    ds_wc: xr.Dataset | None = None,
    elevation_cmap: str = "inferno",
    add_hillshade: bool = True,
    diff_cmap: str = "RdBu",
    elevation_clim: tuple[float, float] | None = None,
    diff_clim: tuple[float, float] = (-10, 10),
    figsize: tuple[float, float] = (17, 11),
) -> plt.Axes:
    """
    Create a panel figure comparing DEM and altimetry elevations.

    Where the first row shows elevation maps over hillshades (if add_hillshade is True) altimetry points will be plotted over the hillshade of the first DEM (if add_hillshade is True)
    The second row shows elevation differences, with the first plot being the Landcover of the scene First DEM provided will be the 'source' DEM where all other elevations will be compared to
    The third row shows the histograms of these differences

    Parameters
    ----------
    dem_list: list
        List of xr.Datasets. Each Dataset should have .elevation and .hillshade variables.
        The first DEM in this list will be the 'source' DEM and will served as a reference for differencing
        e.g.  [dem_1,dem_2,dem_3] will result in diff_1 = dem_2 - dem_1 and diff_2 = dem_3 - dem_1
    gdf_dict: Optional Dict[str, Tuple[gpd.GeoDataFrame, str]]
        Dictionary where keys are subplot names and values are (GeoDataFrame, column_name) pairs.
        column_name denotes elevation values to plot (e.g. h_li for ICESat-2 ATL06)
    ds_wc: Optional xr.Dataset
        WorldCover dataset to plot. If this is not provided, ESA WorldCover is used.
    elevation_cmap: Optional str
        Colormap for elevation. Default: 'inferno'
    add_hillshade: Optional bool
        Whether to plot hillshades under your DEMs. Default True.
        If True, your DEMs should have a .hillshade variable
    diff_cmap: Optional str
        Colormap for elevation differences. Default: 'RdBu'
    elevation_clim: Optional Tuple[float, float]
        Tuple for elevation color limits
    diff_clim: Optional Tuple[float, float]
        Tuple for difference color limits. Default (-10,10)
    figsize: Optional Tuple[int, int]
        Figure size tuple. Will be dynamically adjusted based on number of DEMs and GeoDataFrames
        so you typically don't want to adjust this

    Returns
    ----------
        Dictionary containing the axes objects for each subplot

    Notes
    -----
    All inputs assumed to have the same CRS and to be aligned
    All DEMs must have variable 'elevation'
    A minimum requirement of 2 DEMs in dem_list
    A maximum of 5 total datasets (dem_list + gdf_dict) to be passed
    """
    # TODO: better DEM titles in plots
    # Calculate number of columns for the plot (DEMs + GeoDataFrames if provided)
    n_dems = len(dem_list)
    gdf_keys = list(gdf_dict.keys()) if gdf_dict else []
    n_columns = max(2, n_dems) + len(gdf_keys)
    n_columns = min(n_columns, 5)

    # Dynamically adjust figure size
    figsize = (3 * n_columns, 11)
    mosaic = [
        [f"dem_{i}" for i in range(n_dems)] + gdf_keys,
        ["worldcover"]
        + [f"diff_{i}" for i in range(1, n_dems)]
        + [f"diff_{g}" for g in gdf_keys],
        ["wc_legend"]
        + [f"hist_{i}" for i in range(1, n_dems)]
        + [f"hist_{g}" for g in gdf_keys],
    ]
    gs_kw = {"width_ratios": [1] * len(mosaic[0]), "height_ratios": [2, 2, 1]}
    _, axd = plt.subplot_mosaic(
        mosaic, gridspec_kw=gs_kw, figsize=figsize, layout="constrained"
    )
    reference_ax = axd["dem_0"]
    for key in axd:
        if not key.startswith("hist") and not key.startswith("wc"):
            axd[key].sharex(reference_ax)
            axd[key].sharey(reference_ax)

    # TODO: move this to a top-level helper function, and set aspect based on mid scene latitude
    # or equal if UTM, etc. ?
    def _style_ax(
        ax: plt.Axes,
        facecolor: str | None = None,
        aspect: str | None = None,
    ) -> None:
        """Helper function to style axes (remove ticks and set face color)."""
        ax.set_xticks([])
        ax.set_yticks([])
        if facecolor:
            ax.set_facecolor(facecolor)
        if aspect:
            ax.set_aspect(aspect)

    # Generate WorldCover dataset if not provided
    if ds_wc is None:
        bounds = dem_list[0].rio.bounds()
        gf_search = gpd.GeoDataFrame(
            geometry=[box(*bounds)],
            crs=dem_list[0].rio.crs,
        )
        gf_wc = search(
            dataset="worldcover",
            intersects=gf_search,
            datetime=["2020"],
        )
        # TODO: set requested resolution based on AOI / DEM resolution?
        # Or always keep at native 10m resolution ?
        ds_wc = to_dataset(
            gf_wc,
            bands=["map"],
            aoi=gf_search,
        ).compute()

    # TOP ROW: Elevation maps
    # ---------------------------
    # raster elevs
    for i, dem in enumerate(dem_list):
        dem_key = f"dem_{i}"
        plot_dem(
            dem.elevation,
            axd[dem_key],
            elevation_cmap,
            da_hillshade=dem.hillshade if add_hillshade else None,
            add_colorbar=False,
            vmin=(elevation_clim[0] if elevation_clim else None),
            vmax=(elevation_clim[1] if elevation_clim else None),
            title=dem_key,
        )
        _clear_labels(axd[dem_key])
        _style_ax(axd[dem_key])
        if i == n_columns - 1:  # Add colorbar to the last GeoDataFrame plot
            # images[1] since [0] would be the hillshade
            plt.colorbar(axd[dem_key].images[1], ax=axd[dem_key], label="Elevation (m)")
        # TODO: add scalebar later, allow for lon/lat or meters
        # elif i == 0:
        #     dx = dem.rio.resolution()[0]
        #     scalebar = ScaleBar(
        #         dx,  # meters per pixel from resolution
        #         "m",
        #         length_fraction=0.25,
        #         location="lower right",
        #         box_alpha=0.5,
        #         box_color="white",
        #         color="black",
        #         pad=0.5,
        #     )
        #     axd[dem_key].add_artist(scalebar)

    # point elevs
    # need this for MyPy to handle the chance the gdf_dict is None
    # despite the logic in the loop "if gdf_dict else []"
    if gdf_dict:
        # Plot GeoDataFrames (e.g., IS2 and GEDI)
        for key, (gdf, column) in gdf_dict.items() if gdf_dict else []:
            plot_altimeter_points(
                gdf,
                column,
                ax=axd[key],
                title=key,
                da_hillshade=dem_list[0].hillshade if add_hillshade else None,
                vmin=(elevation_clim[0] if elevation_clim else None),
                vmax=(elevation_clim[1] if elevation_clim else None),
                s=(figsize[0] * figsize[1])
                / n_columns**2,  # Dynamic point size based on figure dimensions
            )
            _style_ax(axd[key])
            if key == list(gdf_dict.keys())[-1]:  # Add colorbar to the last gdf plot
                plt.colorbar(
                    axd[key].collections[0], ax=axd[key], label="Elevation (m)"
                )

    # MIDDLE ROW: Differences
    # ---------------------------
    plot_esa_worldcover(
        ds_wc.isel(time=0),
        ax=axd["worldcover"],
        cax=axd["wc_legend"],
        add_colorbar=True,
        add_labels=False,
    )
    _style_ax(axd["worldcover"])
    axd["worldcover"].set_title("ESA WorldCover 2020")

    # raster diffs
    for i, dem in enumerate(dem_list[1:], start=1):
        diff = get_elev_diff(dem, dem_list[0])
        diff_key = f"diff_{i}"
        if add_hillshade:
            dem_list[0].hillshade.plot.imshow(
                ax=axd[diff_key],
                cmap="gray",
                alpha=1.0,
                add_colorbar=False,
                add_labels=False,
            )
        diff.elev_diff.squeeze().plot.imshow(
            ax=axd[diff_key],
            cmap=diff_cmap,
            add_colorbar=False,
            add_labels=False,
            vmin=diff_clim[0],
            vmax=diff_clim[1],
            alpha=0.5 if add_hillshade else 1.0,
        )
        _style_ax(axd[diff_key], facecolor="black")
        axd[diff_key].set_title(f"dem_{i} minus dem_0")
        if i == n_columns - 1:  # Add colorbar to the last gdf plot
            plt.colorbar(
                axd[diff_key].images[1] if add_hillshade else axd[diff_key].images[0],
                ax=axd[diff_key],
                label="Elevation Difference (m)",
            )
    # gdf diffs
    # need this for MyPy to handle the chance the gdf_dict is None
    # despite the logic in the loop "if gdf_dict else []"
    if gdf_dict:
        for _, (key, (gdf, column)) in enumerate(gdf_dict.items() if gdf_dict else []):
            diff_key = f"diff_{key}"
            if add_hillshade:
                dem_list[0].hillshade.plot.imshow(
                    ax=axd[diff_key],
                    cmap="gray",
                    alpha=1.0,
                    add_colorbar=False,
                    add_labels=False,
                )
            gf_diff = get_elev_diff(gdf, dem_list[0], source_col=column)
            plot_altimeter_points(
                gf_diff,
                "elev_diff",
                axd[diff_key],
                cmap=diff_cmap,
                vmin=diff_clim[0],
                vmax=diff_clim[1],
                title=f"{column} minus dem_0",
                s=(figsize[0] * figsize[1])
                / n_columns**2,  # Dynamic sizing based on figure dimensions
            )
            _style_ax(axd[diff_key], facecolor="black")
            if key == list(gdf_dict.keys())[-1]:  # Add colorbar to the last gdf plot
                plt.colorbar(
                    axd[diff_key].collections[0],
                    ax=axd[diff_key],
                    label="Elevation Difference (m)",
                )

    # BOTTOM ROW: Histograms
    # ---------------------------
    axd["wc_legend"].axis("off")

    # raster hists
    for i, dem in enumerate(dem_list[1:], start=1):
        hist_key = f"hist_{i}"
        plot_diff_hist(
            get_elev_diff(dem, dem_list[0]).elev_diff,
            ax=axd[hist_key],
            dmin=diff_clim[0],
            dmax=diff_clim[1],
        )
        axd[hist_key].set_title("")

    # gdf hists
    for _, (key, (gdf, column)) in enumerate(gdf_dict.items() if gdf_dict else []):
        hist_key = f"hist_{key}"
        diff = get_elev_diff(gdf, dem_list[0], source_col=column)
        plot_diff_hist(
            diff.elev_diff, ax=axd[hist_key], dmin=diff_clim[0], dmax=diff_clim[1]
        )
        axd[hist_key].set_title("")

    return axd


@depends_on_optional("matplotlib")
def boxplot_terrain_diff(
    dem_list: list[xr.Dataset | gpd.GeoDataFrame],
    ax: plt.Axes | None = None,
    terrain_v: str = "slope",
    terrain_groups: np.ndarray | None = None,
    ylims: list[float] | None = None,
    title: str = "Elevation Differences (m) by Terrain",
    ylabel: str = "Elevation Difference (m)",
    elev_col: str | None = None,
) -> plt.Axes:
    """
    Box plots of elevation differences by terrain group for two input elevation datasets.
    (e.g. elevation differences over varying slopes for two DEMs)
    This also shows the counts of the grouped elevation differences

    Parameters
    ----------
    dem_list : list[xr.Dataset | gpd.GeoDataFrame]
        List containing exactly two elevation datasets to compare. The first
        dataset must be a xr.Dataset with an 'elevation' variable and a variable
        corresponding to the terrain_v variable (e.g. 'slope').
    terrain_v : str
        Terrain variable of interest (e.g. 'slope') that exists in first DEM. This
        is what your elevation differences will be grouped by
    terrain_groups : np.ndarray
        Array defining the edges of terrain groups. e.g. if set to np.arange(0, 46, 1),
        the groups will be [0, 1, 2, ..., 45]
    ylims : list
        The y-axis limits for the plot. If None, then the y-axis limits will be
        automatically adjusted to fit the whiskers of each boxplot
    title : str
        The title of the plot. Default is 'Elevation Differences (m) by Terrain'
    ylabel : str
        The yabel of the plot. Default is 'Elevation Difference (m)'
    elev_col : str
        The name of the column containing elevation values to difference in
        the GeoDataFrame if a GeoDataFrame is provided

    Returns
    -------
    plt.Axes
        The matplotlib axes object containing the plot.

    Notes
    -----
    This function assumes that the datasets in dem_list are in the same CRS and aligned
    This function can also work with point geometry GeoDataFrames (e.g. ICESat-2 points)
    If using a GeoDataFrame, you must also provide the elev_col parameter which is the column name containing elevation values you wish to compare
    This function requires there to be EXACTLY 2 datasets in the dem_list.
    The first dataset in dem_list MUST be a xr.Dataset with an 'elevation' variable and the corresponding terrain_v variable (e.g. 'slope')
    """
    if len(dem_list) != 2:
        msg_len = "dem_list must contain exactly two datasets"
        raise ValueError(msg_len)
    # ruff B008 if default is set to np.arange(0, 46, 1)
    if terrain_groups is None:
        msg_groups = "terrain_groups is undefined"
        raise ValueError(msg_groups)
    # difference (second - first) based on type (xr.dataset vs gdf)
    if isinstance(dem_list[1], gpd.GeoDataFrame):
        if elev_col is None:
            msg_col_arg = "elev_col must be provided when using point data"
            raise ValueError(msg_col_arg)
        da_points = dem_list[1].get_coordinates().to_xarray()
        samples = (
            dem_list[0]
            .interp(da_points)
            .drop_vars(["band", "spatial_ref"])
            .to_dataframe()
        )
        dem_list[1]["elev_diff"] = (
            dem_list[1][elev_col].to_numpy() - samples["elevation"].to_numpy()
        )
        dem_list[1][terrain_v] = samples[terrain_v].to_numpy()
        diff_data = dem_list[1]["elev_diff"]
    else:
        diff_data = dem_list[1]["elevation"] - dem_list[0]["elevation"]

    if ax is None:
        _, ax = plt.subplots(figsize=(14, 6))
    box_data = []  # difference data per group
    box_labels = []  # xtick labels for box groups
    counts = []  # group observation counts
    # loop over terrain groups and extract differences by group
    for i in range(len(terrain_groups) - 1):
        b1, b2 = terrain_groups[i], terrain_groups[i + 1]

        if isinstance(dem_list[1], gpd.GeoDataFrame):
            mask = (dem_list[1][terrain_v] > b1) & (dem_list[1][terrain_v] <= b2)
            data = dem_list[1].loc[mask, "elev_diff"].dropna()
        else:
            mask = (dem_list[0][terrain_v] > b1) & (dem_list[0][terrain_v] <= b2)
            data = diff_data.where(mask).to_numpy()
            data = data[~np.isnan(data)]

        # minimum 30 observations per group
        if len(data) >= 30:
            box_data.append(data)
            box_labels.append(f"{b1}-{b2}")
            counts.append(len(data))

    if len(box_data) == 0:
        msg_counts = "No groups satisfied the minimum 30 observation count threshold."
        raise ValueError(msg_counts)

    ax.boxplot(
        box_data,
        orientation="vertical",
        patch_artist=True,
        showfliers=True,
        boxprops={"facecolor": "lightgray", "color": "black"},
        flierprops={"marker": "x", "color": "black", "markersize": 5, "alpha": 0.6},
        medianprops={"color": "black"},
        positions=np.arange(len(box_data)),
    )

    # second axis for observation counts
    ax2 = ax.twinx()
    ax2.plot(np.arange(len(counts)), counts, "o", color="orange", alpha=0.6)
    # dynamic y-ticks for observation counts
    magnitude = 10 ** np.floor(np.log10(max(counts)))
    min_count = np.floor(min(counts) / magnitude) * magnitude
    max_count = np.ceil(max(counts) / magnitude) * magnitude
    ticks = np.linspace(min_count, max_count, 11)
    ax2.set_yticks(ticks)
    ax2.spines["right"].set_color("orange")
    ax2.tick_params(axis="y", colors="orange")
    ax2.set_ylabel("Count", color="orange", fontsize=12)

    # original axis elements
    ax.axhline(0, color="black", linestyle="dashed", linewidth=0.7)
    global_median = np.nanmedian(diff_data.to_numpy())
    ax.axhline(
        global_median,
        color="magenta",
        linestyle="dashed",
        linewidth=0.7,
        label=f"Global Med: {global_median:.2f}m",
        alpha=0.8,
    )
    # Dynamic ylims for the elevation differences
    # Zooms into whiskers extent
    if ylims is None:
        # get whiskers to fit the ylims of the plot
        data_concat = np.concatenate(box_data)
        q1, q3 = np.percentile(data_concat, [25, 75])
        iqr = q3 - q1
        ymin = np.floor(q1 - 1.5 * iqr)
        ymax = np.ceil(q3 + 1.5 * iqr)
        # force include 0 in y axis
        ymin = min(ymin, 0)
        ymax = max(ymax, 0)
        n_ticks = 11
        spacing = max(abs(ymin), abs(ymax)) * 2 / (n_ticks - 1)
        spacing = np.ceil(spacing)
        ymin = np.floor(ymin / spacing) * spacing
        ymax = np.ceil(ymax / spacing) * spacing
        yticks = np.arange(ymin, ymax + spacing, spacing)
        ylims = [ymin, ymax]
        ax.set_yticks(yticks)
    ax.set_ylim(ylims)
    ax.set_xticks(np.arange(len(box_labels)))
    ax.set_xticklabels(box_labels, rotation=45, fontsize=10)
    ax.set_ylabel(ylabel, fontsize=12)
    ax.set_xlabel(terrain_v, fontsize=12)
    ax.set_title(title, fontsize=14)

    lines1, labels1 = ax.get_legend_handles_labels()
    ax.legend(lines1, labels1, loc="best", fontsize=10)

    return ax


# slope wrapper for boxplot_terrain_diff()
@depends_on_optional("matplotlib")
def boxplot_slope(
    dem_list: list[xr.Dataset | gpd.GeoDataFrame],
    ax: plt.Axes | None = None,
    slope_bins: np.ndarray | None = None,
    ylims: list[float] | None = None,
    title: str = "Elevation Differences (m) by Slope",
    ylabel: str = "Elevation Difference (m)",
    elev_col: str | None = None,
) -> plt.Axes:
    """
    Boxplots of elevation differences grouped by elevation values.
    Grouped on the first dem in dem_list, which must have an 'slope' variable

    This is a wrapper around boxplot_terrain_diff() specifically for slope analysis
    with groups from 0-45 degrees in 1 degree increments.

    Parameters
    ----------
    dem_list : list[xr.Dataset | gpd.GeoDataFrame]
        List containing exactly two elevation datasets to compare
    slope_bins : np.ndarray
        Array defining the edges of terrain groups. e.g. if set to np.arange(0, 46, 1),
        the groups will be [0, 1, 2, ..., 45]
        Default is np.arange(0, 46, 1)
    ylims : list[float] | None
        The y-axis limits for the plot
    title : str
        The title of the plot
    ylabel : str
        The ylabel of the plot
    elev_col : str | None
        Column name containing elevation values if using GeoDataFrame
    show : bool
        Whether to display the plot

    Returns
    -------
    plt.Axes
        The matplotlib axes object containing the plot
    """
    if slope_bins is None:
        slope_bins = np.arange(0, 46, 1)
    return boxplot_terrain_diff(
        dem_list=dem_list,
        ax=ax,
        terrain_v="slope",
        terrain_groups=np.arange(0, 46, 1),
        ylims=ylims,
        title=title,
        ylabel=ylabel,
        elev_col=elev_col,
    )


@depends_on_optional("matplotlib")
def boxplot_elevation(
    dem_list: list[xr.Dataset | gpd.GeoDataFrame],
    ax: plt.Axes | None = None,
    elevation_bins: np.ndarray | None = None,
    ylims: list[float] | None = None,
    title: str = "Elevation Differences (m) by Source Elevation",
    ylabel: str = "Elevation Difference (m)",
    elev_col: str | None = None,
) -> plt.Axes:
    """
    Boxplots of elevation differences grouped by elevation values.
    Grouped on the first dem in dem_list, which must have an 'elevation' variable

    This is a wrapper around boxplot_terrain_diff() specifically for elevation analysis
    with groups defined by elevation_bins. If elevation_bins is None, bins are created
    from the min/max elevations of the first DEM in steps of 300m.

    Parameters
    ----------
    dem_list : list[xr.Dataset | gpd.GeoDataFrame]
        List containing exactly two elevation datasets to compare
    elevation_bins : np.ndarray | None
        Array defining the edges of elevation groups
        If None, bins are created from min/max of source DEM in 300m steps
    ylims : list[float] | None
        The y-axis limits for the plot
    title : str
        The title of the plot
    ylabel : str
        The ylabel of the plot
    elev_col : str | None
        Column name containing elevation values if using GeoDataFrame

    Returns
    -------
    plt.Axes
        The matplotlib axes object containing the plot
    """
    if elevation_bins is None:
        elev_min = np.floor(np.nanmin(dem_list[0].elevation) / 100) * 100
        elev_max = np.ceil(np.nanmax(dem_list[0].elevation) / 100) * 100
        elevation_bins = np.arange(elev_min, elev_max + 300, 300)

    return boxplot_terrain_diff(
        dem_list=dem_list,
        ax=ax,
        terrain_v="elevation",
        terrain_groups=elevation_bins,
        ylims=ylims,
        title=title,
        ylabel=ylabel,
        elev_col=elev_col,
    )


@depends_on_optional("matplotlib")
def boxplot_aspect(
    dem_list: list[xr.Dataset | gpd.GeoDataFrame],
    ax: plt.Axes | None = None,
    aspect_bins: np.ndarray | None = None,
    ylims: list[float] | None = None,
    title: str = "Elevation Differences (m) by Source Aspect",
    ylabel: str = "Elevation Difference (m)",
    elev_col: str | None = None,
) -> plt.Axes:
    """
    Boxplots of elevation differences grouped by aspect values.
    Grouped on the first dem in dem_list, which must have an 'aspect' variable

    This is a wrapper around boxplot_terrain_diff() specifically for aspect analysis
    with groups defined by aspect_bins. If aspect_bins is None, bins are created
    from 0-360 degrees in steps of 10 degrees.

    Parameters
    ----------
    dem_list : list[xr.Dataset | gpd.GeoDataFrame]
        List containing exactly two elevation datasets to compare
    aspect_bins : np.ndarray | None
        Array defining the edges of aspect groups
        If None, bins are created from 0-360 in steps of 10 degrees
    ylims : list[float] | None
        The y-axis limits for the plot
    title : str
        The title of the plot
    ylabel : str
        The ylabel of the plot
    elev_col : str | None
        Column name containing elevation values if using GeoDataFrame

    Returns
    -------
    plt.Axes
        The matplotlib axes object containing the plot
    """
    if aspect_bins is None:
        aspect_bins = np.arange(0, 370, 10)

    return boxplot_terrain_diff(
        dem_list=dem_list,
        ax=ax,
        terrain_v="aspect",
        terrain_groups=aspect_bins,
        ylims=ylims,
        title=title,
        ylabel=ylabel,
        elev_col=elev_col,
    )


@depends_on_optional("matplotlib")
def _create_stats_legend(
    ax: plt.Axes, stats_dict: dict[str, float], color_dict: dict[str, str]
) -> None:
    """
    Helper function for hist_esa. Create a unified legend with statistics for elevation difference plots.

    Parameters
    ----------
    ax : plt.Axes
        The matplotlib axes object to add the legend to
    stats_dict : dict[str, float]
        Dictionary containing statistics to display
    color_dict : dict[str, str]
        Dictionary mapping statistic names to colors

    Returns
    -------
    None
        Modifies the input axes object directly
    """
    legend_elements = []

    for label in ["Mean", "Med"]:
        color = color_dict.get(label, "black")
        legend_elements.append(
            plt.Line2D([0], [0], color=color, label=f"{label}: {stats_dict[label]:.2f}")
        )

    for label in ["Std", "NMAD"]:
        legend_elements.append(
            plt.Line2D(
                [0], [0], color="none", label=f"{label}: {stats_dict[label]:.2f}"
            )
        )

    legend_elements.append(
        plt.Line2D([0], [0], color="none", label=f"Count: {int(stats_dict['Count'])}")
    )

    legend = ax.legend(
        handles=legend_elements,
        loc="upper right",
        bbox_to_anchor=(0.98, 0.98),
        fontsize=10,
        facecolor="lightgray",
        edgecolor="none",
    )
    legend.get_frame().set_alpha(0.5)


@depends_on_optional("matplotlib")
def hist_esa(
    dem_list: list[xr.Dataset | gpd.GeoDataFrame],
    elev_col: str | None = None,
    min_count: int = 30,
    color_dict: dict[str, str] | None = None,
) -> plt.Figure:
    """
    Histogram of elevation differences between DEMs or point data, grouped by ESA World Cover 2020 land cover class.

    Parameters
    ----------
    dem_list : list[xr.Dataset | gpd.GeoDataFrame]
        List containing two elevation datasets to compare
    elev_col : str | None
        Column name containing elevation values if using point data
    min_count : int
        Minimum number of points required for a land cover class to be included
    color_dict : dict[str, str]
        Dictionary mapping statistics to colors for plotting

    Returns
    -------
    np.array
        An array containing the matplotlib axes objects
    """
    if len(dem_list) != 2:
        msg_len = "dem_list must contain exactly two datasets"
        raise ValueError(msg_len)
    if color_dict is None:
        color_dict = {
            "Mean": "magenta",
            "Med": "deepskyblue",
            "Std": "black",
            "NMAD": "black",
            "Count": "black",
        }
    # calculate elevation differences if second dataset is a GDF
    if isinstance(dem_list[1], gpd.GeoDataFrame):
        if elev_col is None:
            col_msg = "elev_col must be provided when using point data"
            raise ValueError(col_msg)

        da_points = dem_list[1].get_coordinates().to_xarray()
        samples = (
            dem_list[0]
            .interp(da_points)
            .drop_vars(["band", "spatial_ref"])
            .to_dataframe()
        )
        dem_list[1]["elev_diff"] = (
            dem_list[1][elev_col].to_numpy() - samples["elevation"].to_numpy()
        )
        diff_data = dem_list[1]
    else:  # Both are rasters
        diff_data = dem_list[1]["elevation"] - dem_list[0]["elevation"]

    classmap = WorldCover().classmap

    # Get WorldCover dataset
    bounds = dem_list[0].rio.bounds()
    gf_search = gpd.GeoDataFrame(
        geometry=[box(*bounds)], crs=dem_list[0].rio.crs
    ).to_crs(epsg=4326)
    gf_wc = search(dataset="worldcover", intersects=gf_search, datetime=["2020"])
    # For tiny datasets (e.g. tests), mask=True can fail..
    # TODO: revisit masking here
    ds_wc = to_dataset(gf_wc, bands=["map"], aoi=gf_search)

    ds_wc = ds_wc.rio.reproject(dem_list[0].rio.crs).rio.reproject_match(dem_list[0])

    # plotting logic for raster data (xarray)
    if isinstance(diff_data, xr.DataArray):
        bin_width = 0.25
        bins = np.arange(-30, 30 + bin_width, bin_width)
        filtered_classes = []

        for class_value, class_info in classmap.items():
            class_mask = ds_wc.isel(time=0)["map"].to_numpy() == class_value
            class_mask_da = xr.DataArray(
                class_mask, coords=[diff_data.y, diff_data.x], dims=["y", "x"]
            )
            class_data = diff_data.where(class_mask_da, drop=True)
            class_data_flat = class_data.to_numpy().flatten()
            class_data_flat = class_data_flat[~np.isnan(class_data_flat)]

            if len(class_data_flat) >= min_count:
                filtered_classes.append((class_value, class_info, class_data_flat))

        num_classes = len(filtered_classes)
        ncols, nrows = 2, (num_classes + 1) // 2
        _, axes = plt.subplots(
            nrows,
            ncols,
            figsize=(12, 4 * nrows),
            constrained_layout=True,
            sharex=True,
            sharey=True,
        )

        for idx, (ax, (_class_value, class_info, class_data_flat)) in enumerate(
            zip(axes.ravel(), filtered_classes, strict=False)
        ):
            ax.hist(
                class_data_flat, bins=bins, alpha=0.5, color=class_info["hex"], log=True
            )
            ax.set_title(class_info["description"], fontsize=12)
            ax.axvline(0, color="k", linestyle="dashed", linewidth=1)
            ax.set_xlim(-20, 20)
            # only set ylabel if axis is in first column
            if idx % ncols == 0:
                ax.set_ylabel("Log(Count)")
            else:
                ax.set_ylabel("")

            # calculate statistics
            stats_dict = {
                "Mean": np.mean(class_data_flat),
                "Med": np.median(class_data_flat),
                "Std": np.std(class_data_flat),
                "NMAD": stats.median_abs_deviation(class_data_flat, scale="normal"),
                "Count": len(class_data_flat),
            }

            _create_stats_legend(ax, stats_dict, color_dict)

            # vertical lines for med and mean
            ax.axvline(stats_dict["Med"], color="deepskyblue", lw=0.5)
            ax.axvline(stats_dict["Mean"], color="magenta", lw=0.5)

        plt.suptitle(
            "Elevation Differences by Land Cover (ESA World Cover 2020)", fontsize=14
        )

    # plotting logic if second dataset is a GDF
    elif isinstance(diff_data, gpd.GeoDataFrame):
        # sample worldcover values at points
        land_cover = ds_wc.isel(time=0)["map"].interp(
            x=("points", diff_data.geometry.x.to_numpy()),
            y=("points", diff_data.geometry.y.to_numpy()),
            method="nearest",
        )
        diff_data["land_cover_class"] = land_cover.to_numpy()

        groups = diff_data.groupby("land_cover_class")
        # filter out groups with less than min_count points
        filtered_groups = [
            (label, group) for label, group in groups if len(group) >= min_count
        ]

        num_groups = len(filtered_groups)
        ncols, nrows = 2, (num_groups + 1) // 2
        _, axes = plt.subplots(
            nrows,
            ncols,
            figsize=(12, 4 * nrows),
            constrained_layout=True,
            sharex=True,
            sharey=True,
        )

        for idx, (ax, (label, group)) in enumerate(
            zip(axes.ravel(), filtered_groups, strict=False)
        ):
            color = classmap.get(label, {}).get("hex", "#cccccc")
            ax.hist(
                group["elev_diff"].dropna(), bins=256, alpha=0.5, color=color, log=True
            )
            ax.axvline(0, color="k", linestyle="dashed", linewidth=1)
            ax.set_title(classmap.get(label, {}).get("description", ""), fontsize=12)
            ax.set_xlim(-20, 20)
            # only set ylabel if axis is in first column
            if idx % ncols == 0:
                ax.set_ylabel("Log(Count)")
            else:
                ax.set_ylabel("")

            group_no_nan = group["elev_diff"].dropna()
            stats_dict = {
                "Mean": np.mean(group_no_nan),
                "Med": np.median(group_no_nan),
                "Std": np.std(group_no_nan),
                "NMAD": stats.median_abs_deviation(group_no_nan, scale="normal"),
                "Count": len(group_no_nan),
            }

            _create_stats_legend(ax, stats_dict, color_dict)

            # vertical lines for med and mean
            ax.axvline(stats_dict["Med"], color="deepskyblue", lw=0.5)
            ax.axvline(stats_dict["Mean"], color="magenta", lw=0.5)

        plt.suptitle(
            "Elevation Differences by Land Cover (ESA World Cover 2020)", fontsize=14
        )

    return axes



"""
Spatial metadata for the 3D Elevation Program (3DEP) is available by work unit in the Work Unit Extent Spatial Metadata (WESM).
https://www.usgs.gov/ngp-standards-and-specifications/wesm-data-dictionary
"""

from __future__ import annotations

from importlib import resources
from typing import Any

import pandas as pd
import pyogrio
from cloudpathlib import S3Client
from geopandas import GeoDataFrame, read_file
from pandas import Timedelta, Timestamp
from shapely.geometry import box

from coincident.datasets import usgs
from coincident.overlaps import subset_by_temporal_overlap

# Cloudpath-based S3 client
client = S3Client(no_sign_request=True)
# Geopandas S3 Client
pyogrio.set_gdal_config_options(
    {"AWS_NO_SIGN_REQUEST": True, "GDAL_PAM_ENABLED": False}
)

swath_polygon_csv = resources.files("coincident.search") / "swath_polygons.csv"

defaults = usgs.ThreeDEP()
wesm_gpkg_url = defaults.search


def stacify_column_names(gf: GeoDataFrame) -> GeoDataFrame:
    """
    Renames WESM column names to match STAC (SpatioTemporal Asset Catalog) conventions

    Parameters
    ----------
    gf : GeoDataFrame
        The input GeoDataFrame with columns 'collect_start' and 'collect_end'.

    Returns
    -------
    GeoDataFrame
        The modified GeoDataFrame with renamed columns and additional 'datetime' and 'duration' columns.

    Notes
    -----
        - 'start_datetime': Renamed from 'collect_start'.
        - 'end_datetime': Renamed from 'collect_end'.
        - 'datetime': Midpoint of 'start_datetime' and 'end_datetime'.
        - 'duration': Duration in days between 'start_datetime' and 'end_datetime'.
    """

    name_map = {
        "collect_start": "start_datetime",
        "collect_end": "end_datetime",
    }
    gf = gf.rename(columns=name_map)
    # Add collection name to match STAC (used to identify geodataframe contents in other functions)
    gf["collection"] = "3DEP"
    gf["start_datetime"] = pd.to_datetime(
        gf["start_datetime"], format="ISO8601", errors="coerce"
    )
    gf["end_datetime"] = pd.to_datetime(
        gf["end_datetime"], format="ISO8601", errors="coerce"
    )
    duration = gf.end_datetime - gf.start_datetime
    gf["datetime"] = gf.start_datetime + duration / 2
    gf["dayofyear"] = gf.datetime.dt.dayofyear
    gf["duration"] = duration.dt.days

    return gf


def read_wesm_csv(url: str = wesm_gpkg_url) -> GeoDataFrame:
    """
    Read WESM metadata from a remote CSV file.

    The CSV contains up to date metadata and a mapping of workunit to feature ID (FID)
    in the main GPKG file.

    Parameters
    ----------
    url : str, optional
        The URL or file path to the WESM  index file

    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing the metadata from the CSV file.
    """
    # Cache CSV locally, so subsequent reads are faster
    wesm_csv = client.CloudPath(url.replace(".gpkg", ".csv"))
    df = pd.read_csv(wesm_csv)
    df.index += 1
    df.index.name = "fid"
    return df


def search_bboxes(
    url: str = wesm_gpkg_url,
    intersects: GeoDataFrame | None = None,
    search_start: Timestamp | None = None,
    search_end: Timestamp | None = None,
    # **kwargs: dict[str, Any] | None,
) -> GeoDataFrame:
    """
    Search WESM.gpkg bounding boxes for spatial and temporal overlap

    Parameters
    ----------
    url : str, optional
        The URL or file path to the GeoPackage (GPKG) file.
    intersects : GeoDataFrame, optional
        A GeoDataFrame to spatially intersect with the convex hulls, by default None.
    search_start : Timestamp, optional
        The start time for the temporal search, by default None.
    search_end : Timestamp, optional
        The end time for the temporal search, by default None.

    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing the convex hull geometries and FIDs in EPSG:4326
    """
    # NOTE: much faster to JUST read bboxes, not full geometry or other columns
    sql = "select * from rtree_WESM_geometry"
    df = pyogrio.read_dataframe(
        url, sql=sql
    )  # , use_arrow=True... arrow probably doesn;t matter for <10000 rows?

    bboxes = df.apply(lambda x: box(x.minx, x.miny, x.maxx, x.maxy), axis=1)
    gf = (
        GeoDataFrame(df, geometry=bboxes, crs="EPSG:4326")
        .rename(columns={"id": "fid"})
        .set_index("fid")
    )

    df = read_wesm_csv(url)
    gf = gf.merge(df, right_index=True, left_index=True)

    gf = stacify_column_names(gf)

    if intersects is not None:
        gf = gf[gf.intersects(intersects.geometry.iloc[0])]

    if search_start is not None and search_end is not None:
        gf = subset_by_temporal_overlap(gf, search_start, search_end, temporal_buffer=0)

    return gf


def load_by_fid(
    fids: list[int], url: str = wesm_gpkg_url, **kwargs: dict[str, Any] | None
) -> GeoDataFrame:
    """
    Extracts full geometry from remote WESM GPKG based on Feature ID (FID)

    Parameters
    ----------
    fids : list(int)
        List of Feature IDs (fids) to extract from GPKG
    url : str
        The URL or file path to the GeoPackage (GPKG) file.

    **kwargs : dict[str, Any], optional
        Additional keyword arguments to pass to :meth:`~geopandas.read_file`

    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing the data from the GeoPackage file with column names
        modified by :meth:`stacify_column_names`.
    """
    # Format SQL: # special case for (a) not (a,)
    # Reading a remote WESM by specific FIDs is fast
    query = f"fid in ({fids[0]})" if len(fids) == 1 else f"fid in {*fids,}"

    gf = read_file(
        url,
        where=query,
        **kwargs,
        # mask=mask, # spatial subset intolerably slow for remote GPKG...
        # NOTE: worth additional dependencies for speed?
        # Only faster I think if GPKG is local & large https://github.com/geopandas/pyogrio/issues/252
        # engine='pyogrio',
        # pyarrow=True,
    )

    # For the purposes of footprint polygon search, just ignore datum (NAD83)
    gf = gf.set_crs("EPSG:4326", allow_override=True)

    return stacify_column_names(gf)


def swathtime_to_datetime(
    gf: GeoDataFrame, start_col: str = "START_TIME", end_col: str = "END_TIME"
) -> GeoDataFrame:
    """
    Convert swath GPS time columns to datetimes in a GeoDataFrame

    Parameters
    ----------
    gf : GeoDataFrame
        The GeoDataFrame containing the swath time columns.
    start_col : str, optional
        The name of the column containing the start times, by default 'START_TIME'.
    end_col : str, optional
        The name of the column containing the end times, by default 'END_TIME'.

    Returns
    -------
    GeoDataFrame
        The GeoDataFrame with additional 'start_ 'dayofyear' column.
    """
    gf["collect_start"] = Timestamp("1980-01-06") + gf[start_col].apply(
        lambda x: Timedelta(seconds=x + 1e9)
    )
    gf["collect_end"] = Timestamp("1980-01-06") + gf[end_col].apply(
        lambda x: Timedelta(seconds=x + 1e9)
    )
    # Assuming no flights spanning multiple days / happening at midnight :)

    return stacify_column_names(gf)


def get_swath_polygons(
    workunit: str,
) -> GeoDataFrame:
    """
    Retrieve swath polygons from a remote URL and convert swath time to datetime.

    Parameters
    ----------
    fid : int
        A pandas Series containing WESM feature ID (FID)

    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing the swath polygons with swath time converted to datetime.

    Notes
    -----
        - Swath polygons not available for all workunits
    """
    # NOTE: this CSV comes from crawling WESM USGS/spatial_metadata in S3 bucket
    df = pd.read_csv(swath_polygon_csv)

    try:
        url = df[df.workunit == workunit].swathpolygon_link.iloc[0]
    except Exception as e:
        message = f"No swath polygons found for workunit={workunit}"
        raise ValueError(message) from e

    # Actually read from S3!
    gf = read_file(url)

    gf = swathtime_to_datetime(gf)
    # Swath polygons likely have different CRS (EPSG:6350), so reproject
    return gf.to_crs("EPSG:4326")



"""
Supported datasets

Files are organized by data provider, each of which can contain multiple datasets

ATL03_v6: https://cmr.earthdata.nasa.gov/search/concepts/C2596864127-NSIDC_CPRD
ATL06_v6: https://cmr.earthdata.nasa.gov/search/concepts/C2564427300-NSIDC_ECS
GEDI_2a_v2: https://cmr.earthdata.nasa.gov/search/concepts/C1908348134-LPDAAC_ECS # did this ID change?!
USGS_3DEP: https://cmr.earthdata.nasa.gov/search/concepts/C2021957295-LPCLOUD
WESM:
"""

from __future__ import annotations

from coincident.datasets import (
    csda,
    maxar,
    nasa,
    neon,
    opentopo,
    planetary_computer,
    usgs,
)
from coincident.datasets.general import Dataset

# Convenience mapping of string aliases to supported dataset classes
_datasets = [
    maxar.Stereo(),
    usgs.ThreeDEP(),
    nasa.ICESat2(),
    nasa.GEDI(),
    nasa.ABLVIS2_1(),
    nasa.AFLVIS2_1(),
    planetary_computer.COP30(),
    planetary_computer.WorldCover(),
    csda.TDX(),
    opentopo.NOAA(),
    opentopo.NCALM(),
    neon.NEON(),
]

aliases = [x.alias for x in _datasets]
_alias_to_Dataset = dict(zip(aliases, _datasets, strict=False))

__all__ = ["Dataset", "usgs", "maxar", "nasa", "planetary_computer", "csda", "opentopo"]



"""
coincident: Search and analysis of STV Precursor Coincident Datasets
"""

from __future__ import annotations

from coincident import datasets, io, overlaps, plot, search
from coincident._version import version as __version__

__all__ = ["__version__", "datasets", "search", "overlaps", "io", "plot"]



"""
OpenTopography Datasets (NOAA, NCALM)

NOAA: https://www.opentopography.org/
NCALM: https://www.ncalm.org/
"""

from __future__ import annotations

from dataclasses import dataclass

from coincident.datasets.general import Dataset


@dataclass
class NOAA(Dataset):
    """Essential metadata for OpenTopography NOAA LiDAR"""

    has_stac_api: bool = False
    search: str = "https://portal.opentopography.org/API/otCatalog?productFormat=PointCloud&polygon={coords}&detail=true&outputFormat=json&include_federated=true"
    start: str = "1996-10-09"
    end: str | None = None
    type: str = "lidar"
    alias: str = "noaa"
    provider: str = "opentopography"


@dataclass
class NCALM(Dataset):
    """Essential metadata for OpenTopography NCALM LiDAR"""

    has_stac_api: bool = False
    search: str = "https://portal.opentopography.org/API/otCatalog?productFormat=PointCloud&polygon={coords}&detail=true&outputFormat=json&include_federated=false"
    start: str = "2003-05-15"
    end: str | None = None
    type: str = "lidar"
    alias: str = "ncalm"
    provider: str = "opentopography"



"""
Subset and Sample using Xarray
"""

from __future__ import annotations

import os
from typing import Any

import geopandas as gpd
import odc.stac
import pystac
import rasterio

# NOTE: must import for odc.stac outputs to have .rio accessor
import rioxarray  # noqa: F401
import xarray as xr

from coincident.search.stac import to_pystac_items

# Sets GDAL_DISABLE_READDIR_ON_OPEN to 'EMPTY_DIR' etc.
odc.stac.configure_rio(cloud_defaults=True, VSICURL_PC_URL_SIGNING="YES")

# TODO: investigate performance of GTI versus VRT
cop_to_7912 = "https://raw.githubusercontent.com/uw-cryo/3D_CRS_Transformation_Resources/refs/heads/main/globaldems/COP30_hh_7912.vrt"
nasadem_to_7912 = "https://raw.githubusercontent.com/uw-cryo/3D_CRS_Transformation_Resources/refs/heads/main/globaldems/nasadem_7912.vrt"
threedep_to_7912 = "https://raw.githubusercontent.com/uw-cryo/3D_CRS_Transformation_Resources/refs/heads/main/3dep/USGS_Seamless_DEM_13_7912.vrt"


def load_dem_7912(dataset: str, aoi: gpd.GeoDataFrame) -> xr.DataArray:
    """
    Load a Digital Elevation Model (DEM) dataset and clip it to the area of interest (AOI).
    Reprojection-on-reading to EPSG:7912 is performed by GDAL.

    Parameters
    ----------
    dataset : str
        The name of the DEM dataset to load. Must be one of 'cop30', 'nasadem', or '3dep'.
    aoi : gpd.GeoDataFrame
        Polygon with lon,lat coordinates to which the DEM will be clipped.

    Returns
    -------
    xr.DataArray
        The clipped DEM data as an xarray DataArray.

    Notes
    -----
    - This function currently eagerly pulls all data into memory, so may fail if aoi is large
    """
    # NOTE: Currently only getting data from OpenTopography VRTs, but leaving this in for future
    # if dataset in ["cop30", "nasadem"]:
    #     r = requests.get(
    #         "https://planetarycomputer.microsoft.com/api/sas/v1/token/pcstacitems/items",
    #         timeout=10,
    #     )
    #     token = r.json()["token"]
    #     if dataset == "cop30":
    #         uri = cop_to_7912
    #     elif dataset == "nasadem":
    #         uri = nasadem_to_7912
    #     ENV = rasterio.Env(
    #         GDAL_DISABLE_READDIR_ON_OPEN="EMPTY_DIR",
    #         AZURE_STORAGE_ACCOUNT="pcstacitems",
    #         AZURE_STORAGE_SAS_TOKEN=token,
    #         VSICURL_PC_URL_SIGNING="YES",
    #     )
    if dataset == "cop30":
        uri = cop_to_7912
    elif dataset == "nasadem":
        uri = nasadem_to_7912
    elif dataset == "3dep":
        uri = threedep_to_7912
    else:
        msg = f"Unknown dataset: {dataset}, must be 'cop30', 'nasadem', or '3dep'"
        raise ValueError(msg)

    ENV = rasterio.Env(
        GDAL_DISABLE_READDIR_ON_OPEN="EMPTY_DIR", AWS_NO_SIGN_REQUEST="YES"
    )

    # Pull all data into memory
    with ENV:
        return (
            xr.open_dataarray(uri, engine="rasterio")
            .rio.clip_box(**aoi.bounds)
            .squeeze()
            .load()
        )


def to_dataset(
    gf: gpd.GeoDataFrame,
    bands: list[str] | None = None,
    aoi: gpd.GeoDataFrame | None = None,
    mask: bool = False,
    **kwargs: Any,
) -> xr.DataArray:
    """
    Convert a GeoDataFrame to an xarray DataArray using odc.stac

    Parameters
    ----------
    gf : gpd.GeoDataFrame
        The GeoDataFrame containing the geospatial data.
    aoi : gpd.GeoDataFrame, optional
        GeoDataFrame containing area of interest (e.g. search polygon)
    bands : str, optional
        The asset keys to extract from the STAC items, by default ["data"].
    **kwargs : dict
        Additional keyword arguments passed to `odc.stac.load`.

    Returns
    -------
    xr.Dataset
        The resulting stacked xarray Dataset backed by dask arrays
    """
    if bands is None:
        bands = ["data"]

    items = to_pystac_items(gf)
    ds = odc.stac.load(items, bands=bands, geopolygon=aoi, chunks={}, **kwargs)
    if mask:
        ds = ds.rio.clip(aoi)

    return ds


# NOTE: make this more general to open any single STAC asset?
def open_maxar_browse(item: pystac.Item, overview_level: int = 0) -> xr.DataArray:
    """
    Open a browse image from a STAC item using the specified overview level.

    Parameters
    ----------
    item : pystac.Item
        The STAC item containing the browse image asset.
    overview_level : int, optional
        The overview level to use when opening the image, by default 0.

    Returns
    -------
    xr.DataArray
        The opened browse image as an xarray DataArray.

    Notes
    -----
    The function uses the `rasterio` engine to open the image and sets the
    `GDAL_DISABLE_READDIR_ON_OPEN` and `GDAL_HTTP_HEADERS` environment variables
    for optimized reading and authentication, respectively.
    """

    # href = item.assets['browse']['href'] # accessed via geopandas
    href = item.assets["browse"].href  # accessed via pystac

    env = rasterio.Env(
        GDAL_DISABLE_READDIR_ON_OPEN="EMPTY_DIR",
        GDAL_HTTP_HEADERS=f'MAXAR-API-KEY:{os.environ["MAXAR_API_KEY"]}',
    )
    with env:
        return xr.open_dataarray(
            href,
            engine="rasterio",
            mask_and_scale=False,  # otherwise uint8 -> float32!
            backend_kwargs={"open_kwargs": {"overview_level": overview_level}},
        )



from __future__ import annotations

import warnings
from typing import Any

import geopandas as gpd

# Used to access formatters
from pystac_client.item_search import ItemSearch as _ItemSearch

from coincident.datasets import _alias_to_Dataset
from coincident.datasets.general import Dataset
from coincident.overlaps import subset_by_minimum_area
from coincident.search import neon_api, opentopo_api, stac, wesm

_pystac_client = _ItemSearch("no_url")


def search(
    dataset: str | Dataset,
    intersects: gpd.GeoDataFrame | gpd.GeoSeries | None = None,
    datetime: str | list[str] | None = None,
    # NOTE: change to explicitly stac_kwargs? not sure of proper kwargs type (Optional[dict[str, Any]]?)
    **kwargs: Any,
) -> gpd.GeoDataFrame:
    """
    Perform a search for geospatial data using STAC or non-STAC APIs for a single dataset.

    Parameters
    ----------
    dataset : str or Dataset
        The dataset to search. Can be a string alias or a Dataset object.
    intersects : gpd.GeoDataFrame
        A GeoDataFrame containing a single row, or GeoSeries containing a geometry to restrict the search area.
    datetime : str or None, optional
        The datetime range for the search in ISO 8601 format. If None, no datetime filter is applied.
    **kwargs : Any
        Additional keyword arguments to pass to the search functions.
        see https://pystac-client.readthedocs.io/en/latest/api.html#item-search

    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame containing the search results.

    Raises
    ------
    ValueError
        If the provided dataset alias is not supported.
    """
    # Validate Dataset
    if isinstance(dataset, str):
        try:
            dataset = _alias_to_Dataset[dataset]
        except KeyError as e:
            msg_unsupported = (
                f"{dataset} is not a supported dataset: {_alias_to_Dataset.keys()}"
            )
            raise ValueError(msg_unsupported) from e

    # Validate Datetimes
    _validate_temporal_bounds(dataset, datetime)
    if datetime is not None:
        start, end = _pystac_client._format_datetime(datetime).split("/")
        search_start = (
            gpd.pd.to_datetime(start).tz_localize(None) if start != ".." else None
        )
        search_end = gpd.pd.to_datetime(end).tz_localize(None) if end != ".." else None
    else:
        search_start = None  # or  gpd.pd.Timestamp(dataset.start) ?
        search_end = None  # or gpd.pd.Timestamp.today()?

    # Validate Intersects
    if intersects is not None:
        _validate_spatial_bounds(intersects)

        # NOTE: not very robust, explode() demotes MultiPolygons to single Polygon (seems many GeoJSONs have this)
        # ANd 'exterior' not available for Multipolygons, just
        shapely_geometry = intersects.geometry.explode().iloc[0]
        if (
            not shapely_geometry.exterior.is_ccw
        ):  # Apparently NASA CMR enforces polygon CCW order
            shapely_geometry = shapely_geometry.reverse()
        aoi = _pystac_client._format_intersects(shapely_geometry)  # to JSON geometry
    else:
        if "bbox" not in kwargs and "ids" not in kwargs:
            msg_unconstrained = (
                "Neither `bbox` nor `intersects` provided... search will be global"
            )
            warnings.warn(msg_unconstrained, stacklevel=2)
        aoi = None

    # STAC API Searches
    # NOTE: refactor so that search config lives alongside dataset? easier to add, would need conditionals below
    if dataset.has_stac_api:
        stac_api_kwargs = {**kwargs, **dataset.stac_kwargs}  # NOTE: override order?
        stac_api_kwargs["datetime"] = datetime
        stac_api_kwargs["collections"] = dataset.collections
        stac_api_kwargs["intersects"] = aoi

        if dataset.provider == "maxar":
            # NOTE: not sure how to avoid incompatible type "str | None"; expected "str" for Dataset.attrs
            client = stac.configure_maxar_client(dataset.area_based_calc)  # type: ignore[attr-defined]
            item_collection = stac.search(client, **stac_api_kwargs)
            gf = stac.to_geopandas(item_collection)
            # Client-side reduce to only acquisitions having stereo pairs
            gf = gf.loc[gf.stereo_pair_identifiers.str[0].dropna().index]

        else:
            # NOTE: NASA-CMR-STAC seems to require GET, but default is POST (maxar API w/ GET throws error for large polygons)
            if dataset.provider == "nasa":
                stac_api_kwargs["method"] = "GET"
                client = stac.configure_stac_client(dataset.search)  # type: ignore[arg-type]
            # Generic STAC endpoint w/o additional config
            else:
                client = stac.configure_stac_client(dataset.search)  # type: ignore[arg-type]
            item_collection = stac.search(client, **stac_api_kwargs)

            # Per-dataset munging
            # https://github.com/uw-cryo/coincident/issues/8#issuecomment-2449810481
            if dataset.alias == "tdx":
                # Drop columns with messy schema
                dropcols = [
                    "sceneInfo",
                    "missionInfo",
                    "previewInfo",
                    "imageDataInfo",
                    "generationInfo",
                    "acquisitionInfo",
                    "productVariantInfo",
                ]
                for item in item_collection:
                    for col in dropcols:
                        item.properties.pop(col)

            gf = stac.to_geopandas(item_collection)

    # Non-STAC Searches
    elif dataset.alias == "3dep":
        gf = wesm.search_bboxes(
            intersects=intersects,
            search_start=search_start,
            search_end=search_end,
            **kwargs,
        )

    elif dataset.alias == "neon":
        gf = neon_api.search_bboxes(
            intersects=intersects,
            search_start=search_start,
            search_end=search_end,
        )

    elif dataset.provider == "opentopography":
        gf = opentopo_api.search_opentopo(
            intersects=intersects,
            search_start=search_start,
            search_end=search_end,
            dataset=dataset.alias,
        )

    # Keep track of dataset alias in geodataframe metadata
    # NOTE: attrs not always retrained and not saved to file
    # https://github.com/geopandas/geopandas/issues/3320
    gf.attrs["dataset"] = dataset.alias

    return gf


def _validate_temporal_bounds(
    dataset: Dataset,
    datetime: str | list[str] | None = None,
) -> None:
    """
    Validates the temporal bounds of a dataset against a given datetime range.

    Parameters
    ----------
    dataset : _Dataset
        The dataset object containing start and end attributes.
    datetime : str, list, or None, optional
        The datetime range to validate against the dataset's temporal bounds.
        It can be a single string, a list of strings, or None. The datetime
        range should be in a format supported by pystac_client.

    Raises
    ------
    ValueError
        If the requested search start date is after the dataset's end date or
        if the requested search end date is before the dataset's start date.
    """
    if datetime is not None:
        # Use pystac_client to parse all supported datetime inputs
        start, end = _pystac_client._format_datetime(datetime).split("/")

        if start != ".." and dataset.end is not None:
            search_start = gpd.pd.to_datetime(start)
            dataset_end = gpd.pd.to_datetime(dataset.end, utc=True)
            if search_start > dataset_end:
                message = f"Requested search start {search_start} > {dataset.alias} end date: {dataset_end}"
                raise ValueError(message)

        if end != ".." and dataset.start is not None:
            search_end = gpd.pd.to_datetime(end)
            dataset_start = gpd.pd.to_datetime(dataset.start, utc=True)
            if search_end < dataset_start:
                message = f"Requested search end {search_end} < {dataset.alias} start date: {dataset_start}"
                raise ValueError(message)


def _validate_spatial_bounds(
    intersects: gpd.GeoSeries,
) -> None:
    """
    Validate that the specified area of interest (AOI) is type GeoDataFrame or GeoSeries
    Parameters
    ----------
    dataset : _Dataset
        The dataset to validate.
    intersects : gpd.GeoSeries
        A GeoSeries containing the area of interest (AOI) geometry.
    Raises
    ------
    ValueError
        If the AOI is not type GeoDataFrame or GeoSeries
    """
    if not isinstance(intersects, gpd.GeoDataFrame | gpd.GeoSeries):
        message = f"intersects value must be a GeoDataFrame or GeoSeries, not {type(intersects)}"
        raise ValueError(message)
    if len(intersects) > 1:
        message = "GeoDataFrame contains multiple geometries, search requires a single geometry"
        raise ValueError(message)


def cascading_search(
    primary_dataset: gpd.GeoDataFrame,
    secondary_datasets: list[tuple[str, int]] = [  # noqa: B006
        ("maxar", 14),
        ("icesat", 40),
        ("gedi", 40),
    ],
    min_overlap_area: float = 20,
) -> list[gpd.GeoDataFrame]:
    """
    Perform an cascading search to find overlapping datasets acquired within specific time ranges.

    Secondary datasets are searched based only on spatial overlap areas with previous datasets. In other words, the overlapping area is progressively reduced.

    Temporal buffer is applied as either (datetime-buffer <= acquisition <= datetime+buffer)
     or (start_datetime-buffer <= acquisition <= end_datetime+buffer)

    Parameters
    ----------
    primary_dataset : gpd.GeoDataFrame
        The primary dataset having 'datetime' or 'start_dateteime' and 'end_datetime' columns.

    secondary_datasets : list of tuple, optional
        Each tuple contains the name of the secondary dataset and temporal buffer in days.

    min_overlap_area : int, optional
        The minimum overlap area in km^2. Default is 20.

    Returns
    -------
    list of gpd.GeoDataFrame
        A list of GeoDataFrames containing the search results for each secondary dataset.
    """
    # Do searches on simple geometry, but intersect results with original geometry
    search_geometry = primary_dataset.simplify(0.01)  # or convex_hull?
    detailed_geometry = primary_dataset[["geometry"]]

    if "end_datetime" in primary_dataset.columns:
        start = primary_dataset.start_datetime.iloc[0]
        end = primary_dataset.end_datetime.iloc[0]
    else:
        start = end = primary_dataset.datetime.iloc[0]

    results = []
    for dataset, temporal_buffer in secondary_datasets:
        pad = gpd.pd.Timedelta(days=temporal_buffer)
        date_range = [start - pad, end + pad]

        # Search secondary dataset
        gfs = search(
            dataset=dataset,
            intersects=search_geometry,
            datetime=date_range,
        )

        if dataset == "maxar":
            gfs["stereo_pair_id"] = gfs.stereo_pair_identifiers.str[0]
            gfs = gfs.dissolve(by="stereo_pair_id", as_index=False)

        # Keep track of original footprints
        gfs["original_geometry"] = gfs["geometry"]

        gf_i = gfs.overlay(detailed_geometry, how="intersection")
        gf_i = subset_by_minimum_area(gf_i, min_overlap_area)
        results.append(gf_i)

        # We've refined our search polygon again, so update search GeoDataFrame
        detailed_geometry = gf_i.dissolve()[["geometry"]]
        search_geometry = detailed_geometry.simplify(0.01)

    return results



"""
Supported datasets:

ATL03_v6: https://cmr.earthdata.nasa.gov/search/concepts/C2596864127-NSIDC_CPRD
ATL06_v6: https://cmr.earthdata.nasa.gov/search/concepts/C2564427300-NSIDC_ECS
GEDI_2a_v2: https://cmr.earthdata.nasa.gov/search/concepts/C1908348134-LPDAAC_ECS
USGS_3DEP: https://cmr.earthdata.nasa.gov/search/concepts/C2021957295-LPCLOUD
"""

# from pydantic.dataclasses import dataclass, Field # type: ignore[attr-defined]
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any


# NOTE: default to None for all of these?
@dataclass
class Dataset:
    """Essential metadata for supported datasets"""

    alias: str | None = None  # nickname
    has_stac_api: bool | None = None  # whether or not its a STAC API
    collections: list[str] = field(
        default_factory=list
    )  # STAC collection names of specific datasets
    search: str | None = None  # search API endpoint
    start: str | None = None  # first acquisition date
    end: str | None = None  # last acquisition date (or None if ongoing)
    type: str | None = None  # lidar | stereo | altimeter | sar
    provider: str | None = None  # usgs, maxar, nasa, microsoft, csda
    # Pystac client default limit=100, but seems set by API endpoint as well (nasa cmr-stac=25)
    stac_kwargs: dict[str, Any] = field(default_factory=lambda: {"limit": 1000})



"""
Functions for refining search results based on spatial and temporal overlaps
"""

from __future__ import annotations

from typing import Any

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely.geometry
from shapely.geometry.polygon import orient as _orient


def subset_by_maximum_duration(
    gf: gpd.GeoDataFrame,
    max_duration: int = 60,  # days
) -> gpd.GeoDataFrame:
    """
    Subset a GeoDataFrame by a maximum duration.

    Parameters
    ----------
    gf : _gpd.GeoDataFrame
        The input GeoDataFrame containing a 'duration' column.
    max_duration : int, optional
        The maximum duration in days to filter the GeoDataFrame. Default is 60 days.

    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame filtered to include only rows where the 'duration' is less than or equal to the specified maximum duration.
    """
    max_duration = pd.Timedelta(days=max_duration)
    # NOTE: keep indexes?
    return gf.loc[gf.duration <= max_duration, :]


def subset_by_temporal_overlap(
    gf: gpd.GeoDataFrame,
    start_datetime: pd.Timestamp,
    end_datetime: pd.Timestamp,
    temporal_buffer: int = 14,
) -> gpd.GeoDataFrame:
    """
    Subset GeoDataFrame with time intervals by a single temporal span.

    Parameters
    ----------
    gf : gpd.GeoDataFrame
        The input GeoDataFrame with 'start_datetime' and 'end_datetime' columns
    start_datetime : pd.Timestamp
        The start datetime for the overlap.
    end_datetime : pd.Timestamp
        The end datetime for the overlap.
    temporal_buffer : int, optional
        The buffer in days to apply to the start and end datetimes (default is 14).

    Returns
    -------
    gpd.GeoDataFrame
        A subset of the input GeoDataFrame with rows that overlap temporally with the specified span.
    """
    temporal_buffer = pd.Timedelta(days=temporal_buffer)
    # Apply temporal buffer to collection start and end dates
    buffered_start = gf.start_datetime - temporal_buffer
    buffered_end = gf.end_datetime + temporal_buffer

    results_intervals = gpd.pd.IntervalIndex.from_arrays(
        buffered_start, buffered_end, closed="both"
    )
    search_interval = gpd.pd.Interval(start_datetime, end_datetime, closed="both")
    keep = results_intervals.overlaps(search_interval)  # pylint: disable=no-member

    return gf.loc[keep, :]


def geographic_area(gf: gpd.GeoDataFrame) -> gpd.pd.Series:
    """
    Calculate the geographic area of each polygon in a GeoDataFrame in KM^2

    Parameters
    ----------
    gf : gpd.GeoDataFrame
        A GeoDataFrame containing the geometries for which the area needs to be calculated. The GeoDataFrame
        must have a geographic coordinate system (latitude and longitude).

    Returns
    -------
    pd.Series
        A Pandas Series containing the area of each polygon in the input GeoDataFrame.

    Raises
    ------
    TypeError
        If the GeoDataFrame does not have a geographic coordinate system.

    References
    ----------
    - https://gis.stackexchange.com/questions/413349/calculating-area-of-lat-lon-polygons-without-transformation-using-geopandas
    - https://pyproj4.github.io/pyproj/stable/api/geod.html
    """
    if not gf.crs and gf.crs.is_geographic:
        msg = "geodataframe should have geographic coordinate system"
        raise TypeError(msg)

    geod = gf.crs.get_geod()

    # TODO: sort out numpy typing https://github.com/python/mypy/issues/5480
    def area_calc(geom: shapely.geometry) -> Any:
        if geom.geom_type not in ["MultiPolygon", "Polygon"]:
            return np.nan

        # For MultiPolygon do each separately
        if geom.geom_type == "MultiPolygon":
            return np.sum([area_calc(p) for p in geom.geoms])

        # orient to ensure a counter-clockwise traversal.
        # geometry_area_perimeter returns (area, perimeter)
        return geod.geometry_area_perimeter(_orient(geom, 1))[0]

    return gf.geometry.apply(area_calc)


def subset_by_minimum_area(
    gf: gpd.GeoDataFrame,
    min_area: float = 20.0,  # square kilometers
) -> gpd.GeoDataFrame:
    """
    Subset a GeoDataFrame by a minimum area threshold.

    Parameters
    ----------
    gf : gpd.GeoDataFrame
        The input GeoDataFrame containing geographic features.
    min_area : int, optional
        The minimum area threshold in square kilometers. Features with an area
        less than or equal to this value will be included in the subset. Default is 20.

    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame containing only the features with an area less than or equal
        to the specified minimum area.
    """
    areas = geographic_area(gf) * 1e-6

    return gf.loc[areas >= min_area, :]



"""
Wrapper for various data provider search APIs

Results returned as GeoDataFrame with consistent STAC column headings and best dtypes
"""

from __future__ import annotations

from coincident.search.main import cascading_search, search
from coincident.search.stac import to_geopandas, to_pystac_items

__all__ = ["search", "cascading_search", "to_pystac_items", "to_geopandas"]



from __future__ import annotations

import typing
from importlib.metadata import version

import geopandas as gpd
import pytest
from geopandas.testing import assert_geodataframe_equal
from shapely.geometry import Polygon

import coincident

# Decorate tests requiring internet (slow & flaky)
network = pytest.mark.network

try:
    import maxar_platform.discovery  # noqa: F401

    not_authenticated = False
except:  # noqa: E722
    not_authenticated = True
maxar_authenticated = pytest.mark.skipif(
    not_authenticated, reason="Not authenticated with Maxar API"
)

expected_stac_version = "1.1.0" if version("pystac") >= "1.12" else "1.0.0"


@typing.no_type_check
def test_no_dataset_specified():
    with pytest.raises(
        TypeError, match="missing 1 required positional argument: 'dataset'"
    ):
        coincident.search.search(intersects="-120, 40, -121, 41")


def test_unknown_dataset_specified():
    with pytest.raises(ValueError, match="is not a supported dataset"):
        coincident.search.search(dataset="typo", intersects="-120, 40, -121, 41")


def test_polygon_invalid_type():
    with pytest.raises(
        ValueError, match="intersects value must be a GeoDataFrame or GeoSeries"
    ):
        coincident.search.search(dataset="3dep", intersects="-120, 40, -121, 41")


def test_to_geopandas_empty_search_result():
    with pytest.raises(ValueError, match="ItemCollection is empty"):
        coincident.search.stac.to_geopandas([])


def test_unconstrained_search_warns():
    with pytest.warns(match="Neither `bbox` nor `intersects` provided"):
        coincident.search.search(dataset="tdx")


@network
def test_cascading_search(aoi):
    aoi["datetime"] = gpd.pd.to_datetime("2019-06-12", utc=True)
    pad = 30
    secondary_datasets = [("icesat-2", pad), ("gedi", pad)]
    results = coincident.search.cascading_search(aoi, secondary_datasets)

    expected_min = aoi.datetime.iloc[0] - gpd.pd.Timedelta(days=pad)
    expected_max = aoi.datetime.iloc[0] + gpd.pd.Timedelta(days=pad)
    actual_min = results[0].datetime.min()
    actual_max = results[0].datetime.max()

    assert isinstance(results, list)
    assert len(results) == 2
    assert len(results[0]) == 5
    assert len(results[1]) == 4
    assert actual_min >= expected_min
    assert actual_max <= expected_max


# MAXAR
# =======
@network
@maxar_authenticated
@pytest.mark.filterwarnings("ignore:Server does not conform")
def test_maxar_search(aoi):
    gf = coincident.search.search(
        dataset="maxar",
        intersects=aoi,
        datetime="2023",
        filter="eo:cloud_cover < 20",
    )
    assert gf.shape == (10, 71)
    assert gf.iloc[0].stac_version == expected_stac_version
    assert {"browse", "cloud-cover", "sample-point-set"}.issubset(
        gf.iloc[0].assets.keys()
    )
    assert isinstance(gf.datetime.iloc[0], gpd.pd.Timestamp)
    assert {"id", "dayofyear", "stereo_pair_identifiers", "platform"}.issubset(
        set(gf.columns)
    )


@network
@maxar_authenticated
def test_maxar_large_aoi(large_aoi):
    gf = coincident.search.search(
        dataset="maxar",
        intersects=large_aoi,
        datetime="2023",
        # NOTE: this limits total *mono* results, which may or may not have stereo
        max_items=10,
    )
    assert len(gf) <= 10


@network
@maxar_authenticated
def test_maxar_specific_ids():
    gf = coincident.search.search(
        dataset="maxar", ids=["102001008EC5AC00", "102001008BE9BB00"]
    )
    assert set(gf.id) == {"102001008BE9BB00", "102001008EC5AC00"}


# NASA
# =======
expected_nasa_columns = {
    "assets",
    "bbox",
    "collection",
    "geometry",
    "id",
    "links",
    "stac_extensions",
    "stac_version",
    "type",
    "datetime",
    "end_datetime",
    "start_datetime",
    "dayofyear",
}


@network
def test_icesat2_search(aoi):
    gf = coincident.search.search(
        dataset="icesat-2",
        intersects=aoi,
        datetime="2023",
    )
    actual_columns = set(gf.columns)
    data_assets = list(
        filter(lambda x: x["roles"] == "data", gf.iloc[0].assets.values())
    )
    assert gf.iloc[0].stac_version == expected_stac_version
    assert gf.shape == (25, 13)
    assert actual_columns == expected_nasa_columns
    assert "roles" in gf.iloc[0].assets["browse"]
    assert len(data_assets) == 1
    assert gf.iloc[0].collection.startswith("ATL03")
    assert isinstance(gf.start_datetime.iloc[0], gpd.pd.Timestamp)


@network
def test_gedi_search(aoi):
    gf = coincident.search.search(
        dataset="gedi",
        intersects=aoi,
        datetime="2022",
    )
    actual_columns = set(gf.columns)
    data_assets = list(
        filter(lambda x: x["roles"] == "data", gf.iloc[0].assets.values())
    )
    assert gf.iloc[0].stac_version == expected_stac_version
    assert gf.shape == (33, 13)
    assert actual_columns == expected_nasa_columns
    assert "roles" in gf.iloc[0].assets["browse"]
    assert len(data_assets) == 1
    assert gf.iloc[0].collection.startswith("GEDI02_A")
    assert isinstance(gf.start_datetime.iloc[0], gpd.pd.Timestamp)


# NASA/CSDA
# =======
@network
def test_tdx_search(aoi):
    gf = coincident.search.search(
        dataset="tdx", intersects=aoi, datetime=["2009", "2020"]
    )
    expected_columns = {
        "assets",
        "bbox",
        "collection",
        "geometry",
        "id",
        "links",
        "stac_extensions",
        "stac_version",
        "type",
        "constellation",
        "datetime",
        "end_datetime",
        "platform",
        "sar:center_frequency",
        "sar:frequency_band",
        "sar:instrument_mode",
        "sar:looks_azimuth",
        "sar:looks_range",
        "sar:polarizations",
        "sar:product_type",
        "sar:resolution_azimuth",
        "sar:resolution_range",
        "start_datetime",
        "dayofyear",
    }
    actual_columns = set(gf.columns)
    # NOTE: I think technically assets should have 'roles' not 'role' key...
    data_assets = list(
        filter(lambda x: x["role"] == "data", gf.iloc[0].assets.values())
    )
    assert gf.iloc[0].stac_version == expected_stac_version
    assert gf.shape == (48, 24)
    assert actual_columns == expected_columns
    assert gf["sar:product_type"].unique() == "SSC"
    assert len(data_assets) >= 1


# MS PLANETARY COMPUTER
# =======
@network
def test_cop30_search(aoi):
    gf = coincident.search.search(dataset="cop30", intersects=aoi)
    expected_columns = {
        "assets",
        "bbox",
        "collection",
        "geometry",
        "id",
        "links",
        "stac_extensions",
        "stac_version",
        "type",
        "datetime",
        "gsd",
        "platform",
        "proj:code",
        "proj:shape",
        "proj:transform",
        "dayofyear",
    }
    actual_columns = set(gf.columns)
    assert gf.iloc[0].stac_version == expected_stac_version
    assert gf.shape == (4, 16)
    assert actual_columns == expected_columns
    assert "data" in gf.iloc[0].assets
    assert gf.iloc[0]["proj:code"] == "EPSG:4326"


@network
def test_worldcover_search(aoi):
    gf = coincident.search.search(dataset="worldcover", intersects=aoi, datetime="2020")
    # NOTE: could test these separately...
    stac_columns = {
        "assets",
        "bbox",
        "collection",
        "geometry",
        "id",
        "links",
        "stac_extensions",
        "stac_version",
        "type",
        "datetime",
    }
    additional_columns = {
        "created",
        "description",
        "end_datetime",
        "esa_worldcover:product_tile",
        "esa_worldcover:product_version",
        "grid:code",
        "instruments",
        "mission",
        "platform",
        "proj:code",
        "start_datetime",
    }
    coincident_columns = {"dayofyear"}
    expected_columns = stac_columns | additional_columns | coincident_columns
    actual_columns = set(gf.columns)
    assert gf.iloc[0].stac_version == expected_stac_version
    assert gf.shape == (4, 22)
    assert actual_columns == expected_columns
    assert "map" in gf.iloc[0].assets


@network
def test_round_trip_parquet(aoi):
    outpath = "/tmp/search_results.parquet"
    A = coincident.search.search(dataset="cop30", intersects=aoi)
    A.to_parquet(outpath)
    B = gpd.read_parquet(outpath)
    assert_geodataframe_equal(A, B)


# USGS
# =======
@network
def test_wesm_search(aoi):
    gf = coincident.search.search(
        dataset="3dep",
        intersects=aoi,
    )
    expected_columns = {
        "workunit",
        "project",
        "start_datetime",
        "end_datetime",
        "dayofyear",
        "duration",
        "p_method",
    }
    actual_columns = set(gf.columns)
    assert gf.shape == (5, 37)
    assert expected_columns.issubset(actual_columns)


# NOTE can take ~10s on wifi... maybe only run this on demand...
@network
def test_get_swath_polygon():
    gf = coincident.search.wesm.get_swath_polygons("CO_CameronPkFire_1_2021")
    expected_columns = {
        "start_datetime",
        "end_datetime",
        "collection",
        "datetime",
        "dayofyear",
        "duration",
    }
    actual_columns = set(gf.columns)
    assert isinstance(gf, gpd.GeoDataFrame)
    assert gf.shape == (51, 16)
    assert expected_columns.issubset(actual_columns)
    assert isinstance(gf.datetime.iloc[0], gpd.pd.Timestamp)


@network
def test_swath_polygon_not_found():
    with pytest.raises(ValueError, match="No swath polygons found for workunit="):
        coincident.search.wesm.get_swath_polygons("AL_SWCentral_1_B22")


# opentopo (NCALM and NOAA)
# =======
@network
def test_noaa_search(bathy_aoi):
    gf = coincident.search.search(
        dataset="noaa", intersects=bathy_aoi, datetime=["2019-01-01", "2023-12-31"]
    )
    assert gf.shape == (2, 5)
    assert all(
        col in gf.columns
        for col in ["id", "title", "start_datetime", "end_datetime", "geometry"]
    )
    assert all(isinstance(geom, Polygon) for geom in gf["geometry"])


@network
def test_ncalm_search(large_aoi):
    gf = coincident.search.search(
        dataset="ncalm", intersects=large_aoi, datetime=["2019-01-01", "2023-12-31"]
    )
    assert gf.shape == (6, 5)
    assert all(
        col in gf.columns
        for col in ["id", "title", "start_datetime", "end_datetime", "geometry"]
    )
    assert all(isinstance(geom, Polygon) for geom in gf["geometry"])


# NEON
# TODO: add a test for provisional NEON datasets
# =======
@network
@pytest.mark.filterwarnings("ignore:Geometry is in a geographic CRS:UserWarning")
def test_neon_search():
    intersects = gpd.read_file(
        "https://raw.githubusercontent.com/unitedstates/districts/refs/heads/gh-pages/states/MA/shape.geojson"
    )
    gf = coincident.search.search(
        dataset="neon", intersects=intersects, datetime=["2019"]
    )
    assert gf.shape == (2, 6)
    assert all(
        col in gf.columns
        for col in [
            "id",
            "title",
            "start_datetime",
            "end_datetime",
            "product_url",
            "geometry",
        ]
    )
    assert all(isinstance(geom, Polygon) for geom in gf["geometry"])
